import numpy as np
from sklearn.linear_model import LogisticRegression


class SafetyLogisticRegression(LogisticRegression):

    def predict_proba(self, X):
        return np.nan_to_num(super().predict_proba(X))



from itertools import chain
from typing import List

import matplotlib.pyplot as plt
import numpy as np
from deap.gp import Terminal
from deap.tools import HallOfFame
from scipy.spatial import KDTree
from scipy.spatial.distance import correlation
from scipy.stats import spearmanr
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.metrics import r2_score
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import StandardScaler
from xgboost import XGBRegressor

from evolutionary_forest.component.evaluation import quick_evaluate
from evolutionary_forest.multigene_gp import result_calculation, MultipleGeneGP
from evolutionary_forest.sklearn_utils import cross_val_predict


class SurrogateModel():
    """
    Some useful functions for building a surrogate model
    Pre-selection strategy:
    1. simple-task: A simple surrogate task
    2. surrogate-model:
    """

    def __init__(self):
        self.hof: HallOfFame = None
        self.pre_selection: bool = None
        self.X: np.ndarray = None
        self.y: np.ndarray = None
        self.n_pop: int = None
        self.original_features: bool = None
        self.pop: List[MultipleGeneGP] = None

    def feature_surrogate_model(self):
        # construct a surrogate model to predict whether a feature is useful or not
        # negative samples
        data = []
        label = []
        for ind in sorted(self.pop, key=lambda x: x.fitness.wvalues[0])[:self.n_pop // 2]:
            funcs = self.toolbox.compile(ind)
            for func, coef in zip(funcs, ind.coef):
                Yp = result_calculation([func], self.X[:5], self.original_features)
                if coef < 0.001:
                    # label.append(1)
                    pass
                else:
                    data.append(Yp.flatten())
                    label.append(0)
                # label.append(ind.fitness.wvalues[0])
        for ind in sorted(self.pop, key=lambda x: x.fitness.wvalues[0])[self.n_pop // 2:]:
            funcs = self.toolbox.compile(ind)
            for func, coef in zip(funcs, ind.coef):
                Yp = result_calculation([func], self.X[:5], self.original_features)
                if coef < 0.001:
                    # label.append(1)
                    pass
                else:
                    data.append(Yp.flatten())
                    label.append(2)
        data = np.array(data)
        label = np.array(label)
        cross_val_predict(KNeighborsClassifier(n_neighbors=5, metric=correlation), data, label, cv=5)

    def get_features(self, sample, individuals, flatten=False):
        all_Yp = []
        for individual in individuals:
            func = self.toolbox.compile(individual)
            Yp = result_calculation(func, self.X[sample], self.original_features)
            if flatten:
                all_Yp.append(Yp.flatten())
                # all_Yp.append((cross_val_predict(LinearRegression(), Yp, self.y[sample]) - self.y[sample]) ** 2)
            else:
                all_Yp.append(Yp)
        return np.array(all_Yp)

    def pre_selection_individuals(self, parents, offspring, pop_size):
        predicted_values = self.pre_selection_score(offspring, parents)
        final_offspring = []
        if self.pre_selection == 'model-size-medium':
            start_index = pop_size // 2
            index = np.argsort(-1 * predicted_values)[start_index:start_index + pop_size]
        else:
            index = np.argsort(-1 * predicted_values)[:pop_size]
        for i in index:
            final_offspring.append(offspring[i])
        return final_offspring, predicted_values[index]

    def pre_selection_score(self, offspring, parents):
        # larger is better
        if self.pre_selection == 'surrogate-model':
            predicted_values = self.calculate_score_by_surrogate_model(offspring, parents)
        elif self.pre_selection == 'simple-task':
            predicted_values = self.calculate_score_by_statistical_methods(offspring, parents)
        elif self.pre_selection in ['model-size', 'model-size-medium']:
            predicted_values = self.calculate_score_by_model_size(offspring, parents)
        elif self.pre_selection == 'diversity':
            plot = False
            # Get the most hard 20 data points
            all_values = np.array([p.case_values for p in parents])
            sample = np.sum(all_values, axis=0).argsort()[-20:]

            pset = self.toolbox.expr.keywords['pset']

            def quick_get_features(individuals):
                all_Yp = []
                for individual in individuals:
                    Yp = []
                    for gene in individual.gene:
                        input_data = self.X[sample]
                        temp_result = quick_evaluate(gene, pset, input_data)
                        if np.size(temp_result) < len(input_data):
                            temp_result = np.full(len(input_data), temp_result)
                        Yp.append(temp_result)
                    Yp = np.array(Yp).T
                    all_Yp.append(Yp.flatten())
                return np.array(all_Yp)

            offspring_features = quick_get_features(offspring)

            offspring_features = StandardScaler().fit_transform(offspring_features)
            pca = PCA(n_components=10)
            offspring_features = pca.fit_transform(offspring_features)
            if plot:
                plt.scatter(offspring_features[:, 0], offspring_features[:, 1])
            kmeans = KMeans(n_clusters=self.n_pop)
            kmeans.fit(offspring_features)

            predicted_values = np.zeros(len(offspring))
            kd = KDTree(offspring_features)
            nearest_inds = []
            # Get the nearest point for each cluster center
            for c in kmeans.cluster_centers_:
                dist, ind = kd.query(c)
                nearest_inds.append(ind)
            predicted_values[np.array(nearest_inds)] = 1
            if plot:
                plt.scatter(offspring_features[np.array(nearest_inds), 0],
                            offspring_features[np.array(nearest_inds), 1], marker="s", color='red')
                plt.scatter(offspring_features[np.arange(0, 100), 0],
                            offspring_features[np.arange(0, 100), 1], marker=".", color='green')
                plt.show()
        else:
            raise Exception
        return predicted_values

    def calculate_score_by_model_size(self, offspring, parents):
        return np.array([-1 * np.mean([len(y) for y in x.gene]) for x in offspring])

    def calculate_score_by_statistical_methods(self, offspring, parents):
        all_values = np.array([p.case_values for p in parents])
        sample = np.sum(all_values, axis=0).argsort()[-50:]
        offspring_features = self.get_features(sample, offspring, False)
        all_score = []
        for f in offspring_features:
            pipe = self.get_base_model()
            y_pred = cross_val_predict(pipe, f, self.y[sample], cv=3).flatten()
            score = r2_score(self.y[sample], y_pred)
            all_score.append(score)
        return np.array(all_score)

    def calculate_score_by_surrogate_model(self, offspring, parents):
        def spearman(ya, yb):
            correlation = spearmanr(ya, yb)[0]
            return correlation

        sample = np.random.randint(0, len(self.y), size=10)
        # select individuals based on a surrogate model
        parent_features = self.get_features(sample, chain(parents, self.hof), True)
        target = np.array([p.fitness.wvalues[0] for p in chain(parents, self.hof)])

        surrogate_model = XGBRegressor(n_jobs=1, objective='rank:pairwise')
        surrogate_model.fit(parent_features, target)
        offspring_features = self.get_features(sample, offspring, True)
        predicted_values = surrogate_model.predict(offspring_features)

        # scoring = make_scorer(spearman)
        # print('CV', np.mean(cross_val_score(surrogate_model, parent_features, target, cv=3, scoring=scoring)))
        # print('CV', np.mean(cross_val_score(surrogate_model, parent_features, target, cv=3)))
        return predicted_values

    def get_genotype_features(self, pop, get_label=False):
        data = []
        for p in pop:
            height = np.mean([g.height for g in p.gene])
            size = np.mean([len(g) for g in p.gene])
            primitives_dict = {t.name: 0 for t in self.pset.primitives[object]}
            terminals_dict = {t.name: 0 for t in filter(lambda x: isinstance(x, Terminal),
                                                        self.pset.terminals[object])}
            for g in p.gene:
                for x in g:
                    if x.name in primitives_dict:
                        primitives_dict[x.name] += 1
                    if x.name in terminals_dict:
                        terminals_dict[x.name] += 1
            used_variables = len(list(filter(lambda x: x > 0, terminals_dict.values())))
            frequency = list(map(lambda x: x / size, list(primitives_dict.values()) + list(terminals_dict.values())))
            if hasattr(p, 'parent_fitness'):
                parent_fitness = [max(p.parent_fitness), min(p.parent_fitness), np.mean(p.parent_fitness)]
            else:
                parent_fitness = [0, 0, 0]
            data.append([height, size, used_variables] + frequency + parent_fitness)
        if get_label:
            label = []
            for p in pop:
                label.append(p.fitness.wvalues[0])
            return np.array(data), np.array(label)
        else:
            return np.array(data)

    def surrogate_model_construction(self, pop):
        # construct a surrogate model
        data, label = self.get_genotype_features(pop, get_label=True)
        surrogate_model = ExtraTreesRegressor(n_estimators=100, n_jobs=1)
        surrogate_model.fit(data, label)
        # print(np.mean(cross_val_score(surrogate_model, data, label,
        #                               scoring=make_scorer(lambda x, y: spearmanr(x, y)[0]))))
        return surrogate_model



import copy
import random
from operator import attrgetter
from types import SimpleNamespace
import math
import numpy as np
import seaborn as sns
from deap.tools import selRandom, selTournament
from matplotlib import pyplot as plt
from numba import njit
from numpy.linalg import LinAlgError
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA, KernelPCA
from sklearn.manifold import Isomap, LocallyLinearEmbedding, TSNE, SpectralEmbedding
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from skorch.callbacks import EarlyStopping
from torch import optim
from torch.nn import MSELoss
from umap import UMAP
from evolutionary_forest.model.VAE import NeuralNetTransformer, VAE
from evolutionary_forest.multigene_gp import MultipleGeneGP
from evolutionary_forest.utils import efficient_deepcopy


# @njit(cache=True)
def batch_tournament_selection(individuals, k, tournsize=8, batch_size=8, fit_attr="fitness"):
    fitness_cases_num = len(individuals[0].case_values)
    idx_cases_batch = np.arange(0, fitness_cases_num)
    np.random.shuffle(idx_cases_batch)
    _batches = np.array_split(idx_cases_batch, max(fitness_cases_num // batch_size, 1))
    batch_ids = list(range(0, len(_batches)))
    assert len(_batches[0]) >= batch_size or fitness_cases_num < batch_size

    chosen = []
    while len(chosen) < k:
        batches: list = copy.deepcopy(batch_ids)
        while len(batches) > 0 and len(chosen) < k:
            idx_candidates = selRandom(individuals, tournsize)
            cand_fitness_for_this_batch = []
            for idx in idx_candidates:
                cand_fitness_for_this_batch.append(np.mean(idx.case_values[_batches[batches[0]]]))
            idx_winner = np.argmin(cand_fitness_for_this_batch)
            winner = idx_candidates[idx_winner]
            chosen.append(winner)
            batches.pop(0)
    return chosen


@njit(cache=True)
def selAutomaticEpsilonLexicaseNumba(case_values, fit_weights, k):
    selected_individuals = []
    avg_cases = 0

    for i in range(k):
        candidates = list(range(len(case_values)))
        cases = np.arange(len(case_values[0]))
        np.random.shuffle(cases)

        while len(cases) > 0 and len(candidates) > 1:
            errors_for_this_case = np.array([case_values[x][cases[0]] for x in candidates])
            median_val = np.median(errors_for_this_case)
            median_absolute_deviation = np.median(np.array([abs(x - median_val) for x in errors_for_this_case]))
            if fit_weights > 0:
                best_val_for_case = np.max(errors_for_this_case)
                min_val_to_survive = best_val_for_case - median_absolute_deviation
                candidates = list([x for x in candidates if case_values[x][cases[0]] >= min_val_to_survive])
            else:
                best_val_for_case = np.min(errors_for_this_case)
                max_val_to_survive = best_val_for_case + median_absolute_deviation
                candidates = list([x for x in candidates if case_values[x][cases[0]] <= max_val_to_survive])
            cases = np.delete(cases, 0)
        avg_cases = (avg_cases * i + (len(case_values[0]) - len(cases))) / (i + 1)
        selected_individuals.append(np.random.choice(np.array(candidates)))
    return selected_individuals, avg_cases


# @timeit
def selAutomaticEpsilonLexicaseFast(individuals, k, return_avg_cases=False):
    fit_weights = individuals[0].fitness.weights[0]
    case_values = np.array([ind.case_values for ind in individuals])
    index, avg_cases = selAutomaticEpsilonLexicaseNumba(case_values, fit_weights, k)
    selected_individuals = [individuals[i] for i in index]
    if return_avg_cases:
        return selected_individuals, avg_cases
    else:
        return selected_individuals


def selRandomPlus(individuals, k, fit_attr="fitness"):
    # Select Top-2 individuals based on a random projected fitness value
    weight = np.random.random(len(individuals[0].case_values))
    return sorted(individuals, key=lambda x: np.sum(x.case_values * weight))[:k]


def selKnockout(individuals, k, version='O', auto_case=False):
    """
    version:
    'O' represents only the one individual each round
    'S' represents only the multiple individuals each round
    """
    # np.unique([x.fitness.wvalues for x in final_pool],return_counts=True)
    cases = np.array(list(range(len(individuals[0].case_values))))
    final_pool = []
    number_of_cases = math.ceil(np.log2(len(individuals)))
    while len(final_pool) < k:
        random.shuffle(individuals)
        if auto_case:
            np.random.shuffle(cases)
            split_cases = np.array_split(cases, number_of_cases)
        pool = list(range(len(individuals)))
        round = 0
        while (version == 'O' and len(pool) > 1) or (version == 'S' and len(pool) > 2):
            intermediate_pool = []
            for i in range(0, len(pool), 2):
                if auto_case:
                    sample_index = split_cases[round]
                else:
                    sample_index = random.choice(cases)
                if i + 1 >= len(pool) or \
                    (np.sum(individuals[pool[i]].case_values[sample_index]) <=
                     np.sum(individuals[pool[i + 1]].case_values[sample_index])):
                    intermediate_pool.append(pool[i])
                else:
                    intermediate_pool.append(pool[i + 1])
            pool = intermediate_pool
            round += 1
        for p in pool:
            final_pool.append(individuals[p])
    return final_pool


def selBagging(individuals, k, fit_attr="fitness"):
    # Select Top-2 individuals based on bagging samples
    sample_indices = np.arange(len(individuals[0].case_values))
    bootstrap_indices = np.random.choice(sample_indices,
                                         size=sample_indices.shape[0],
                                         replace=True)
    return sorted(individuals, key=lambda x: np.sum(x.case_values[bootstrap_indices]))[:k]


def selTournamentNovelty(individuals, k, tournsize, fit_attr="fitness"):
    # Traditional Tournament: Select based on R2 score
    # Novelty Tournament: Select based on R2 score and Novelty score
    chosen = []
    for i in range(k):
        aspirants = selRandom(individuals, tournsize)
        chosen.append(min(aspirants, key=lambda x: np.sum(x.case_values)))
    return chosen


def selDoubleRound(individuals, k, count=5, base_operator='Lexicase', tournsize=3,
                   parsimonious=False):
    # First Round: Traditional Selection Operator
    # Second Round: Crossover
    selected_individuals = []
    for _ in range(k):
        if base_operator == 'Lexicase':
            inds = selAutomaticEpsilonLexicaseFast(individuals, count)
        elif base_operator == 'Tournament':
            inds = selTournamentNovelty(individuals, count, tournsize=tournsize)
        elif base_operator == 'Random':
            inds = selRandom(individuals, count)
        else:
            raise Exception
        base_ind = efficient_deepcopy(inds[0])
        for id in range(1, count):
            ind = inds[id]
            for i, g in enumerate(ind.gene):
                if (parsimonious and count * base_ind.coef[i] < ind.coef[i]) or \
                    (not parsimonious and base_ind.coef[i] < ind.coef[i]):
                    base_ind.coef[i] = ind.coef[i]
                    base_ind.gene[i] = copy.deepcopy(g)
            ## Only replace the worst feature by the best feature
            # min_index = min(range(len(ind.gene)), key=lambda x: base_ind.coef[x])
            # max_index = max(range(len(ind.gene)), key=lambda x: ind.coef[x])
            # if base_ind.coef[min_index] < ind.coef[max_index]:
            #     base_ind.coef[min_index] = ind.coef[max_index]
            #     base_ind.gene[min_index] = copy.deepcopy(ind.gene[max_index])
        del base_ind.fitness.values
        selected_individuals.append(base_ind)
    return selected_individuals


# @timeit
def selAutomaticEpsilonLexicase(individuals, k):
    selected_individuals = []

    for i in range(k):
        candidates = individuals
        cases = list(range(len(individuals[0].case_values)))
        random.shuffle(cases)
        fit_weights = individuals[0].fitness.weights[0]

        while len(cases) > 0 and len(candidates) > 1:
            errors_for_this_case = [x.case_values[cases[0]] for x in candidates]
            median_val = np.median(errors_for_this_case)
            median_absolute_deviation = np.median([abs(x - median_val) for x in errors_for_this_case])
            if fit_weights > 0:
                best_val_for_case = np.max(errors_for_this_case)
                min_val_to_survive = best_val_for_case - median_absolute_deviation
                candidates = list([x for x in candidates if x.case_values[cases[0]] >= min_val_to_survive])
            else:
                best_val_for_case = np.min(errors_for_this_case)
                max_val_to_survive = best_val_for_case + median_absolute_deviation
                candidates = list([x for x in candidates if x.case_values[cases[0]] <= max_val_to_survive])
            cases.pop(0)
        # print('used cases', len(individuals[0].case_values) - len(cases))
        selected_individuals.append(random.choice(candidates))
    return selected_individuals


def selEpsilonLexicase(individuals, k, epsilon):
    selected_individuals = []

    for i in range(k):
        fit_weights = individuals[0].fitness.weights[0]
        candidates = individuals
        cases = list(range(len(individuals[0].case_values)))
        random.shuffle(cases)

        while len(cases) > 0 and len(candidates) > 1:
            if fit_weights > 0:
                best_val_for_case = max([x.case_values[cases[0]] for x in candidates])
                min_val_to_survive_case = best_val_for_case - epsilon
                candidates = list([x for x in candidates if x.case_values[cases[0]] >= min_val_to_survive_case])
            else:
                best_val_for_case = min([x.case_values[cases[0]] for x in candidates])
                max_val_to_survive_case = best_val_for_case + epsilon
                candidates = list([x for x in candidates if x.case_values[cases[0]] <= max_val_to_survive_case])

            cases.pop(0)

        selected_individuals.append(random.choice(candidates))
        # print('used cases', len(individuals[0].case_values) - len(cases))
        # print('remaining candidates', len(candidates))
    return selected_individuals


def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (rho, phi)


def selMAPEliteClustering(individuals, old_list, map_elite_parameter, n_clusters=None):
    """
    :param individuals:
    :param old_list:
    :param map_elite_parameter:
    :param n_clusters:
    :return:
    elite_map: Selected individuals
    pop_pool: Store all individuals
    """
    if len(old_list) < len(individuals):
        map_dict = {}
        for id, ind in enumerate(individuals + old_list):
            map_dict[id] = ind
        return map_dict, list(map_dict.values())
    pop_size = len(individuals)
    # The individual pool may larger than the population size.
    # pool_ratio = map_elite_parameter.get('pool_ratio', 5)
    # pop_pool = list(sorted(individuals + old_list, key=lambda x: x.fitness.wvalues[0], reverse=True))
    # pop_pool = list({tuple(x.predicted_values): x for x in pop_pool}.values())
    # pop_pool = pop_pool[:pool_ratio * pop_size]
    # individuals = pop_pool
    individuals = individuals + old_list
    case_values = np.array([ind.predicted_values for ind in individuals])
    offspring_features = case_values

    fitness_ratio = map_elite_parameter.get('fitness_ratio', 0.2)
    mean_fitness = np.quantile([ind.fitness.wvalues[0] for ind in individuals], fitness_ratio)
    well_individual_values = np.array([ind.predicted_values for ind in
                                       filter(lambda x: x.fitness.wvalues[0] >= mean_fitness, individuals)])
    pca_dimension = map_elite_parameter.get('pca_dimension', 2)
    pca_dimension = min([pca_dimension, well_individual_values.shape[0], well_individual_values.shape[1]])

    # It is very important to find the reason of the singular value error
    try:
        pca = Pipeline(
            [('Scaler', StandardScaler()),
             ('PCA', PCA(n_components=pca_dimension))],
        ).fit(well_individual_values)
        offspring_features = pca.transform(offspring_features)
    except LinAlgError:
        pca = Pipeline(
            [('Scaler', StandardScaler()),
             ('PCA', PCA(n_components=pca_dimension, svd_solver='randomized'))],
        ).fit(well_individual_values)
        offspring_features = pca.transform(offspring_features)
    n_clusters = pop_size if n_clusters is None else n_clusters
    kmeans = Pipeline(
        [('Scaler', StandardScaler()),
         ('K-Means', KMeans(n_clusters=n_clusters))],
    )
    labels = kmeans.fit_predict(offspring_features)

    key_metric = map_elite_parameter['key_metric']
    map_dict = {}

    # Get the nearest point for each cluster center
    def calculate_size(ind: MultipleGeneGP):
        return np.sum([len(g) for g in ind.gene])

    for label, id in zip(labels, range(len(labels))):
        if label not in map_dict:
            map_dict[label] = individuals[id]
        elif key_metric == 'fitness' and individuals[id].fitness.wvalues >= map_dict[label].fitness.wvalues:
            map_dict[label] = individuals[id]
        elif key_metric == 'size' and calculate_size(individuals[id]) <= calculate_size(map_dict[label]):
            map_dict[label] = individuals[id]
    return map_dict, list(map_dict.values())


def selMAPElites(individuals, old_map, map_elite_parameter: dict, target: np.ndarray = None,
                 data_augmentation=0):
    """
    """
    fitness_ratio = map_elite_parameter.get('fitness_ratio', 0.2)
    map_size = map_elite_parameter.get('map_size', 10)
    pca_dimension = map_elite_parameter.get('pca_dimension', 2)
    reduction_method = map_elite_parameter.get('reduction_method', 'PCA')

    plot = False
    individuals = individuals + list(old_map.values())
    individuals = list({tuple(x.predicted_values): x for x in individuals}.values())
    mean_fitness = np.quantile([ind.fitness.wvalues[0] for ind in individuals], fitness_ratio)
    # update a MAP
    map_dict = {}
    case_values = np.array([ind.predicted_values for ind in individuals])
    semantic_matrix = case_values
    well_individuals = [ind for ind in filter(lambda x: x.fitness.wvalues[0] >= mean_fitness, individuals)]
    if data_augmentation > 0:
        original_data = np.array([ind.predicted_values for ind in well_individuals])
        well_individual_values = np.concatenate([
            (1 - data_augmentation) * original_data + data_augmentation * target,
            (data_augmentation - 1) * original_data + (2 - data_augmentation) * target,
        ], axis=0)
    else:
        well_individual_values = np.array([ind.predicted_values for ind in well_individuals])

    semantic_matrix -= target
    well_individual_values -= target
    scaler = StandardScaler(with_mean=False).fit(well_individual_values)
    semantic_matrix = scaler.transform(semantic_matrix)
    well_individual_values = scaler.transform(well_individual_values)
    if data_augmentation > 0:
        original_data -= target
        original_data = scaler.transform(original_data)

    def common_tsne(metric='euclidean', init='pca'):
        return TSNE(n_components=pca_dimension, perplexity=30 if len(well_individual_values) >= 50 else 5,
                    metric=metric, init=init)

    try:
        if reduction_method == 'PCA':
            pca = PCA(n_components=pca_dimension)
        elif reduction_method == 'KPCA(POLY)':
            pca = KernelPCA(n_components=pca_dimension, kernel='poly')
        elif reduction_method == 'KPCA(RBF)':
            pca = KernelPCA(n_components=pca_dimension, kernel='rbf')
        elif reduction_method == 'KPCA(SIGMOID)':
            pca = KernelPCA(n_components=pca_dimension, kernel='sigmoid')
        elif reduction_method == 'KPCA(COSINE)':
            pca = KernelPCA(n_components=pca_dimension, kernel='cosine')
        elif reduction_method == 'KPCA(COSINE)+TSNE':
            pca = Pipeline([('PCA-Preprocess', KernelPCA(kernel='cosine')),
                            ('PCA', common_tsne())])
        elif reduction_method == 'TSNE(Cosine)':
            pca = common_tsne(metric='cosine')
        elif reduction_method == 'Isomap(Cosine)':
            pca = Isomap(metric='cosine')
        elif reduction_method == 'Isomap':
            pca = Isomap(n_components=pca_dimension)
        elif reduction_method == 'LLE':
            pca = LocallyLinearEmbedding(n_components=pca_dimension)
        elif reduction_method == 'MLLE':
            pca = LocallyLinearEmbedding(n_components=pca_dimension, method='modified')
        elif reduction_method == 'HLLE':
            pca = LocallyLinearEmbedding(n_neighbors=10, n_components=pca_dimension, method='hessian')
        elif reduction_method == 'LTSA':
            pca = LocallyLinearEmbedding(n_components=pca_dimension, method='ltsa')
        elif reduction_method == 'TSNE':
            pca = common_tsne()
        elif reduction_method == 'PCA-TSNE':
            pca = Pipeline([('PCA-Preprocess', PCA(n_components=0.99)),
                            ('PCA', common_tsne())])
        elif reduction_method == 'UMAP':
            pca = UMAP(n_components=pca_dimension)
        elif reduction_method == 'SpectralEmbedding':
            pca = SpectralEmbedding(n_components=pca_dimension)
        elif reduction_method == 'Beta-VAE':
            pca = NeuralNetTransformer(VAE,
                                       criterion=MSELoss(),
                                       optimizer=optim.Adam,
                                       module__input_unit=semantic_matrix.shape[1],
                                       max_epochs=2000,
                                       callbacks=[EarlyStopping(patience=50)],
                                       verbose=False)
        else:
            print('Unsupported method', reduction_method)
            raise Exception
        pca.fit(well_individual_values.astype(np.float32))
    except Exception as e:
        if reduction_method == 'PCA':
            pca = PCA(n_components=pca_dimension, svd_solver='randomized')
        elif reduction_method == 'PCA-TSNE':
            pca = Pipeline([('PCA-Preprocess', PCA(0.99, svd_solver='randomized')),
                            ('PCA', common_tsne(init='random'))])
        elif reduction_method == 'KPCA(COSINE)+TSNE':
            pca = Pipeline([('PCA-Preprocess', KernelPCA(kernel='cosine')),
                            ('PCA', common_tsne(init='random'))])
        elif reduction_method == 'TSNE(Cosine)':
            pca = common_tsne(metric='cosine', init='random')
        elif reduction_method == 'LLE':
            pca = LocallyLinearEmbedding(eigen_solver='dense', n_components=pca_dimension)
        elif reduction_method == 'MLLE':
            pca = LocallyLinearEmbedding(eigen_solver='dense', n_components=pca_dimension,
                                         method='modified')
        elif reduction_method == 'HLLE':
            pca = LocallyLinearEmbedding(eigen_solver='dense', n_neighbors=10, n_components=pca_dimension,
                                         method='hessian')
        elif reduction_method == 'LTSA':
            pca = LocallyLinearEmbedding(eigen_solver='dense', n_components=pca_dimension, method='ltsa')
        elif reduction_method == 'TSNE':
            pca = TSNE(n_components=pca_dimension, perplexity=30 if len(well_individual_values) >= 50 else 5,
                       init='random')
        else:
            print('Unsupported reduction algorithm', reduction_method)
            raise e
        pca.fit(well_individual_values.astype(np.float32))
    if (isinstance(pca, Pipeline) and pca._can_transform()) or \
        (not isinstance(pca, Pipeline) and hasattr(pca, 'transform')):
        if data_augmentation > 0:
            semantic_matrix = pca.transform(original_data.astype(np.float32))
        else:
            semantic_matrix = pca.transform(well_individual_values.astype(np.float32))
    else:
        if isinstance(pca, Pipeline):
            semantic_matrix = np.copy(pca[-1].embedding_)
        else:
            semantic_matrix = np.copy(pca.embedding_)
    semantic_matrix = semantic_matrix[:, :pca_dimension + 1]

    for id in range(semantic_matrix.shape[1]):
        bins = np.linspace(semantic_matrix[:, id].min(), semantic_matrix[:, id].max(), map_size + 1)
        semantic_matrix[:, id] = np.digitize(semantic_matrix[:, id], bins)
        data = semantic_matrix[:, id]
        # for the boundary, it should belong to the boundary class
        data[data == map_size + 1] = map_size
        assert np.all(data <= map_size)

    # pool = individuals if pca._can_transform() else well_individuals
    pool = well_individuals
    for id in range(len(pool)):
        case = tuple(int(x) for x in semantic_matrix[id])
        if case not in map_dict:
            map_dict[case] = pool[id]
        elif pool[id].fitness.wvalues >= map_dict[case].fitness.wvalues:
            map_dict[case] = pool[id]

    if plot:
        if (isinstance(pca, Pipeline) and pca._can_transform()) or \
            (not isinstance(pca, Pipeline) and hasattr(pca, 'transform')):
            plot_matrix = pca.transform(well_individual_values)
        else:
            plot_matrix = pca['PCA'].embedding_
        plt.scatter(plot_matrix[:-1, 0], plot_matrix[:-1, 1])
        plt.scatter(plot_matrix[-1, 0], plot_matrix[-1, 1],
                    marker="*", c='red')
        plt.show()

    if plot:
        map_value = np.zeros((map_size, map_size))
        for k, v in map_dict.items():
            map_value[int(k[0]) - 1][int(k[1]) - 1] = v.fitness.wvalues[0]
        sns.heatmap(map_value, linewidth=0.5)
        plt.show()
    return map_dict


def selMaxAngleSelection(individuals: list, k, target: np.ndarray, unique=False):
    """
    A variant of ADS in Qi Chen's paper
    :param individuals: Individuals from the current population
    :param k: Number of individuals to be selected
    :param target: The target semantics
    :return: Parent individuals
    """
    assert k % 2 == 0
    individual_list = []
    for _ in range(k):
        if len(individuals) < 2:
            break
        all_case_values = np.array([ind.predicted_values - target for ind in individuals])
        parent_a = selTournament(individuals, 1, tournsize=7)[0]
        best_one = np.argmin(cosine_similarity(parent_a.case_values.reshape(1, -1), all_case_values)[0])
        parent_b = individuals[best_one]
        if parent_a == parent_b:
            break
        individual_list.append(parent_a)
        individual_list.append(parent_b)
        if unique:
            individuals.remove(parent_a)
            individuals.remove(parent_b)
    return individual_list


def selAngleDrivenSelection(individuals, k, target: np.ndarray, unique=False):
    """
    Original ADS in Qi Chen's paper
    :param individuals: Individuals from the current population
    :param k: Number of individuals to be selected
    :param target: The target semantics
    :return: Parent individuals
    """
    assert k % 2 == 0
    individual_list = []
    for _ in range(k):
        if len(individuals) < 2:
            break
        parent_a = selTournament(individuals, 1, tournsize=7)[0]
        parent_b = None
        n_trails = 10
        threshold_angle = 0
        best_angle = 1
        for i in range(n_trails):
            candidate = selTournament(individuals, 1, tournsize=7)[0]
            a_values = parent_a.case_values - target
            b_values = candidate.case_values - target
            similarity = cosine_similarity(a_values.reshape(1, -1), b_values.reshape(1, -1))[0][0]
            if similarity < threshold_angle:
                parent_b = candidate
                break
            else:
                if similarity < best_angle:
                    best_angle = similarity
                    parent_b = candidate
        if parent_b is None:
            break
        individual_list.append(parent_a)
        individual_list.append(parent_b)
        if unique:
            individuals.remove(parent_a)
            individuals.remove(parent_b)
    return individual_list


def selGPED(individuals, k, phi=None, rho=None):
    def selTournamentPlus(individuals, k, tournsize):
        """
        Select individuals based on case values
        """
        chosen = []
        for i in range(k):
            if len(individuals) > tournsize:
                aspirants = selRandom(individuals, tournsize)
            else:
                aspirants = individuals
            chosen.append(np.argmin(np.sum(aspirants, axis=1)))
        return chosen

    assert k % 2 == 0, 'k must be an even integer'
    inds = []
    case_values = np.array([ind.case_values for ind in individuals])
    if phi is None:
        phi, rho = get_semantic_matrix(case_values)

    for _ in range(k // 2):
        index = selTournamentPlus(case_values, 1, 7)
        parent_a = index[0]
        r = random.random()
        parent_b = None
        indexes = np.arange(0, len(phi))
        if r < 0.2:
            # high, opposite direction
            if phi[parent_a] <= 4:
                tmp_indexes = indexes[phi == (phi[parent_a] + 4)]
                if len(tmp_indexes) > 0:
                    parent_b = selTournamentPlus(case_values[tmp_indexes], 1, 7)[0]
                    parent_b = tmp_indexes[parent_b]
            else:
                tmp_indexes = indexes[phi == (phi[parent_a] - 4)]
                if len(tmp_indexes) > 0:
                    parent_b = selTournamentPlus(case_values[tmp_indexes], 1, 7)[0]
                    parent_b = tmp_indexes[parent_b]
        if r < 0.4 and parent_b is None:
            # median
            tmp_indexes = indexes[(phi != phi[parent_a]) & (rho != rho[parent_a])]
            if len(tmp_indexes) > 0:
                parent_b = selTournamentPlus(case_values[tmp_indexes], 1, 7)[0]
                parent_b = tmp_indexes[parent_b]
        if r < 0.8 and parent_b is None:
            # standard
            tmp_indexes = indexes[(phi != phi[parent_a]) | (rho != rho[parent_a])]
            if len(tmp_indexes) > 0:
                parent_b = selTournamentPlus(case_values[tmp_indexes], 1, 7)[0]
                parent_b = tmp_indexes[parent_b]
        if parent_b is None:
            # low
            tmp_indexes = indexes[(phi == phi[parent_a])]
            if len(tmp_indexes) > 0:
                parent_b = selTournamentPlus(case_values[tmp_indexes], 1, 7)[0]
                parent_b = tmp_indexes[parent_b]
        inds.append(individuals[parent_a])
        inds.append(individuals[parent_b])
    return inds


def get_semantic_matrix(case_values):
    semantic_matrix = case_values
    try:
        semantic_matrix = PCA(2).fit_transform(semantic_matrix)
    except:
        semantic_matrix = PCA(2, svd_solver='randomized').fit_transform(semantic_matrix)
    rho, phi = cart2pol(semantic_matrix[:, 0], semantic_matrix[:, 1])
    bins = np.linspace(rho.min(), rho.max(), 10)
    rho = np.digitize(rho, bins)
    assert np.all(rho <= 10) & np.all(rho > 0)
    bins = np.linspace(phi.min(), phi.max(), 8)
    phi = np.digitize(phi, bins)
    assert np.all(phi <= 8) & np.all(phi > 0)
    return phi, rho


def selAutomaticEpsilonLexicaseK(individuals, k):
    candidates = individuals
    cases = list(range(len(individuals[0].case_values)))
    random.shuffle(cases)
    fit_weights = individuals[0].fitness.weights[0]

    while len(cases) > 0 and len(candidates) > k:
        errors_for_this_case = [x.case_values[cases[0]] for x in candidates]
        median_val = np.median(errors_for_this_case)
        median_absolute_deviation = np.median([abs(x - median_val) for x in errors_for_this_case])
        if fit_weights > 0:
            best_val_for_case = np.max(errors_for_this_case)
            min_val_to_survive = best_val_for_case - median_absolute_deviation
            c = list([x for x in candidates if x.case_values[cases[0]] >= min_val_to_survive])
            if len(c) < k:
                break
            candidates = c
        else:
            best_val_for_case = np.min(errors_for_this_case)
            max_val_to_survive = best_val_for_case + median_absolute_deviation
            c = list([x for x in candidates if x.case_values[cases[0]] <= max_val_to_survive])
            if len(c) < k:
                break
            candidates = c
        cases.pop(0)
    return random.sample(candidates, k)


def selTournamentPlus(individuals, k, tournsize):
    """
    Select individuals based on the sum of case values
    :param individuals: population
    :param k: number of offspring
    :param tournsize: tournament size
    :return:
    """
    chosen = []
    for i in range(k):
        aspirants = selRandom(individuals, tournsize)
        chosen.append(min(aspirants, key=lambda x: np.sum(x.case_values)))
    return chosen


def selHybrid(individuals, k):
    if 'LogisticRegression' in individuals[0].base_model:
        return selAutomaticEpsilonLexicaseFast(individuals, k)
    elif 'DT' in individuals[0].base_model:
        return selTournament(individuals, k, tournsize=7)
    else:
        raise Exception


def selRoulette(individuals, k, fit_attr="fitness"):
    """
    It should be weighted values
    """
    s_inds = sorted(individuals, key=attrgetter(fit_attr), reverse=True)
    sum_fits = sum(getattr(ind, fit_attr).wvalues[0] for ind in individuals)
    chosen = []
    for i in range(k):
        u = random.random() * sum_fits
        sum_ = 0
        for ind in s_inds:
            sum_ += getattr(ind, fit_attr).wvalues[0]
            if sum_ > u:
                chosen.append(ind)
                break

    return chosen


if __name__ == '__main__':
    class MockIndividual():
        def __init__(self):
            self.case_values = np.random.randint(0, 2, 5)
            self.fitness = SimpleNamespace()
            self.fitness.weights = (1,)
            pass


    individuals = [MockIndividual() for _ in range(20)]
    pop = selAutomaticEpsilonLexicase(individuals, 1)
    print(pop)



import array
import random

import numpy
import numpy as np
import pyade
from pyade import *
from deap import base
from deap import creator
from deap import tools


class EnsembleSelection():
    def __init__(self, NDIM=10, generation=10, population_size=50, evaluation_function=None, verbose=False):
        # Problem dimension
        self.generation = generation
        self.population_size = population_size
        self.NDIM = NDIM

        if not hasattr(creator, "EnsembleFitnessMin"):
            creator.create("EnsembleFitnessMin", base.Fitness, weights=(1.0,))
            creator.create("EnsembleIndividual", array.array, typecode='d', fitness=creator.EnsembleFitnessMin)

        toolbox = base.Toolbox()
        toolbox.register("attr_float", random.uniform, -1, 1)
        toolbox.register("individual", tools.initRepeat, creator.EnsembleIndividual, toolbox.attr_float, NDIM)
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)
        toolbox.register("select", tools.selRandom, k=3)
        toolbox.register("evaluate", evaluation_function)
        self.toolbox = toolbox
        self.verbose = verbose

    def run(self, initial_vector=None):
        toolbox = self.toolbox
        # Differential evolution parameters
        CR = 0.25
        F = 1
        MU = self.population_size
        NGEN = self.generation

        # ensure no worse than last iteration
        # first NDIM individuals are inherited from the previous generation
        pop = toolbox.population(n=MU - 2)
        pop += [creator.EnsembleIndividual(np.linspace(-1, 1, self.NDIM))]
        pop += [creator.EnsembleIndividual(initial_vector)]
        hof = tools.HallOfFame(1)
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)

        logbook = tools.Logbook()
        logbook.header = "gen", "evals", "std", "min", "avg", "max"

        # Evaluate the individuals
        fitnesses = toolbox.map(toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit

        record = stats.compile(pop)
        logbook.record(gen=0, evals=len(pop), **record)
        if self.verbose:
            print(logbook.stream)

        for g in range(1, NGEN):
            for k, agent in enumerate(pop):
                a, b, c = toolbox.select(pop)
                y = toolbox.clone(agent)
                index = random.randrange(self.NDIM)
                for i, value in enumerate(agent):
                    if i == index or random.random() < CR:
                        y[i] = a[i] + F * (b[i] - c[i])
                y.fitness.values = toolbox.evaluate(y)
                if y.fitness > agent.fitness:
                    pop[k] = y
            hof.update(pop)
            record = stats.compile(pop)
            logbook.record(gen=g, evals=len(pop), **record)
            if self.verbose:
                print(logbook.stream)

        # print("Best individual is ", hof[0], hof[0].fitness.values[0])
        return hof[0]


class EnsembleSelectionADE():
    def __init__(self, NDIM=10, generation=10, population_size=50, evaluation_function=None, verbose=False,
                 de_algorithm='IL-SHADE'):
        # You may want to use a variable so its easier to change it if we want
        algorithm = {
            'IL-SHADE': pyade.ilshade,
            'DE': pyade.de,
            'JADE': pyade.jade,
            'MPEDE': pyade.mpede,
            'JSO': pyade.jso,
        }[de_algorithm]

        # We get default parameters for a problem with two variables
        params = algorithm.get_default_params(dim=NDIM)

        # We define the boundaries of the variables
        params['bounds'] = np.array([[-1, 1]] * NDIM)

        # We indicate the function we want to minimize
        params['func'] = evaluation_function
        params['population_size'] = population_size
        params['max_evals'] = generation * population_size
        self.algorithm = algorithm
        self.params = params

    def run(self, initial_vector=None):
        # We run the algorithm and obtain the results
        solution, fitness = self.algorithm.apply(**self.params)
        print('OOB', fitness)
        return solution


if __name__ == "__main__":
    e = EnsembleSelection()
    e.run()



import enum
from abc import ABCMeta

import numpy as np
import pandas as pd
from category_encoders import TargetEncoder
from category_encoders.utils import BaseEncoder
from sklearn import clone
from sklearn.base import TransformerMixin, BaseEstimator


class DimensionTransformer(TransformerMixin, BaseEstimator):
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.reshape((-1, 1))

class TargetEncoderNumpy(TargetEncoder):

    def fit(self, X, y=None, **kwargs):
        X=pd.DataFrame(X).astype(str).sum(axis=1).to_frame()
        return super().fit(X, y, **kwargs)

    def transform(self, X, y=None, override_return_df=False):
        X=pd.DataFrame(X).astype(str).sum(axis=1).to_frame()
        return super().transform(X, y, override_return_df)


def make_class(function, arity=None, parameters=None):
    if parameters is None:
        parameters = {}

    class FeatureTransformer(TransformerMixin):
        def __init__(self, *args):
            if isinstance(function, type):
                self.function = function(**parameters)
            else:
                self.function = function
            self.args = args

        def fit(self, X, y=None, **fit_params):
            output = []
            for arg in self.args:
                if isinstance(arg, (int, float, enum.Enum)):
                    output.append(self.decision_function(X, arg))
                else:
                    output.append(arg.fit_transform(X, y).flatten())
            if isinstance(self.function, (BaseEstimator, BaseEncoder)):
                if len(output) == 1:
                    output = np.array(output).reshape(-1, 1)
                else:
                    output = np.array(output).T
                self.function.fit(output, y)
            return self

        def decision_function(self, X, x):
            if isinstance(x, int) or isinstance(x, float):
                return np.full(X.shape[0], x)
            elif isinstance(x, enum.Enum):
                return X[:, x.value]
            elif hasattr(x, 'predict'):
                # inner regressors
                return x.predict(X)
            else:
                return x.transform(X).flatten()

        def transform(self, X):
            if isinstance(self.function, (BaseEstimator, BaseEncoder)):
                output = [self.decision_function(X, x) for x in self.args]
                if len(output) == 1:
                    output = np.array(output).reshape(-1, 1)
                else:
                    output = np.array(output).T
                if hasattr(self.function, 'predict'):
                    output = self.function.predict(output)
                else:
                    output = self.function.transform(output).flatten()
            else:
                output = self.function(*[self.decision_function(X, x) for x in self.args])
            return np.reshape(output, (-1, 1))

    if isinstance(function, ABCMeta):
        FeatureTransformer.__name__ = function.__name__
    else:
        FeatureTransformer.__name__ = function.__name__
    if arity != None:
        FeatureTransformer.__name__ = FeatureTransformer.__name__ + f'_{arity}'
    return FeatureTransformer


class DummyTransformer(TransformerMixin, BaseEstimator):
    """Dummy transformer to output the first column."""

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X[:, 0]



import numpy as np
from category_encoders import OrdinalEncoder, OneHotEncoder
from category_encoders import TargetEncoder, LeaveOneOutEncoder, BinaryEncoder, CatBoostEncoder, PolynomialEncoder, \
    BackwardDifferenceEncoder, CountEncoder, HashingEncoder, HelmertEncoder, JamesSteinEncoder, \
    MEstimateEncoder, QuantileEncoder, SumEncoder, SummaryEncoder
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_selector as selector
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler


class GeneralFeature:
    pass


class CategoricalFeature:
    pass


class BooleanFeature:
    pass


class NumericalFeature:
    pass


class RandomSeed:
    pass


def check_if_all_integers(x):
    "check a pandas.Series is made of all integers."
    return all(float(i).is_integer() for i in np.unique(x))


def type_detection(X):
    input_arr = []
    for x in range(X.shape[1]):
        v = X[:, x]
        if len(np.unique(v)) == 2:
            input_arr.append(BooleanFeature)
        elif len(np.unique(v)) <= 10 and check_if_all_integers(v):
            input_arr.append(CategoricalFeature)
        else:
            input_arr.append(NumericalFeature)
    return input_arr


class FeatureTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, encoding_scheme=None):
        self.encoding_scheme = encoding_scheme

    def fit(self, X, y=None, **fit_params):
        # # Feature Transformation based on different types of features
        # types = type_detection(X)
        # num_cols = []
        # bool_cols = []
        # cat_cols = []
        # for i, t in enumerate(types):
        #     if t == NumericalFeature:
        #         num_cols.append(i)
        #     elif t == BooleanFeature:
        #         bool_cols.append(i)
        #     else:
        #         cat_cols.append(i)

        encoder = {
            'TargetEncoder': TargetEncoder(),
            'LeaveOneOutEncoder': LeaveOneOutEncoder(),
            'BinaryEncoder': BinaryEncoder(),
            'CatBoostEncoder': CatBoostEncoder(),
            'PolynomialEncoder': PolynomialEncoder(),
            'BackwardDifferenceEncoder': BackwardDifferenceEncoder(),
            'CountEncoder': CountEncoder(),
            'HashingEncoder': HashingEncoder(),
            'HelmertEncoder': HelmertEncoder(),
            'JamesSteinEncoder': JamesSteinEncoder(),
            'MEstimateEncoder': MEstimateEncoder(),
            'OneHotEncoder': OneHotEncoder(),
            'OrdinalEncoder': OrdinalEncoder(),
            'QuantileEncoder': QuantileEncoder(),
            'SumEncoder': SumEncoder(),
            'SummaryEncoder': SummaryEncoder(),
            # 'GLMMEncoder': GLMMEncoder(), # Unavailable
        }[self.encoding_scheme]
        num_transformer = Pipeline([
            ('Imputer', SimpleImputer(strategy='mean')),
            ('Encoder', StandardScaler()),
        ])
        cat_transformer = Pipeline(
            [('Imputer', SimpleImputer(strategy='constant', fill_value='missing')),
             ('Encoder', encoder),
             ('Scaler', StandardScaler())]
        )
        transformer = ColumnTransformer(
            transformers=[
                ('num', num_transformer, selector(dtype_include='float64')),
                ('cat', cat_transformer, selector(dtype_include='category')),
                ('obj', cat_transformer, selector(dtype_include='object')),
            ])

        self.x_scaler = transformer
        self.x_scaler.fit(X, y)
        return self

    def transform(self, X):
        return self.x_scaler.transform(X)


if __name__ == '__main__':
    pass



import copy
import itertools
import random
import sys
from functools import wraps
from inspect import isclass
from typing import Callable

import numpy as np
from deap import base
from deap.gp import PrimitiveTree, compile, cxOnePoint, mutUniform, mutShrink, mutInsert, cxOnePointLeafBiased, \
    PrimitiveSet
from deap.tools import selTournament, selRandom
from scipy.special import softmax
from scipy.stats import pearsonr
from tpot.base import TPOTBase

from evolutionary_forest.component.primitives import individual_to_tuple
from evolutionary_forest.component.syntax_tools import TransformerTool


def selTournamentFeature(individuals, k, tournsize):
    chosen = []
    for i in range(k):
        aspirants = selRandom(individuals, tournsize)
        chosen.append(max(aspirants, key=lambda x: x[1]))
    return chosen


class FitnessMin(base.Fitness):
    weights = (-1.0,)


class MultipleGeneGP():
    def __init__(self, content, gene_num, tpot_model: TPOTBase = None, partition_scheme=None,
                 base_model_list=None, number_of_register=0, active_gene_num=0, intron_probability=0):
        self.gene = []
        self.fitness = FitnessMin()
        self.lgp_mode = False
        # Some more basic features
        self.gene_num = gene_num
        self.active_gene_num = active_gene_num
        # This flag is only used for controlling the mutation and crossover
        for i in range(self.gene_num):
            pset = content.keywords['pset']
            if isinstance(pset, MultiplePrimitiveSet):
                self.lgp_mode = True
                self.gene.append(PrimitiveTree(content(pset=pset.pset_list[i])))
            else:
                self.gene.append(PrimitiveTree(content()))
        if tpot_model != None:
            self.base_model = tpot_model._toolbox.individual()
        if base_model_list != None:
            self.base_model = random.choice(base_model_list.split(','))
        self.dynamic_leaf_size = random.randint(1, 10)
        self.dynamic_regularization = 1
        if partition_scheme is not None:
            # self.partition_scheme = np.random.randint(0, partition_scheme[1] , partition_scheme[0])
            # initialize partition scheme by decision tree
            if callable(partition_scheme):
                self.partition_scheme = partition_scheme()
            else:
                self.partition_scheme = partition_scheme
        else:
            self.partition_scheme = None
        self.number_of_register = number_of_register
        # self-adaptive evolution (for Lasso)
        self.parameters = {
            'ActiveGene': np.random.randn(self.gene_num) < intron_probability,
            'Register': np.random.randint(0, self.number_of_register, self.gene_num),
            'Lasso': np.random.uniform(-5, -2, 1)[0],
        }

    def random_select_index(self):
        return random.randint(0, self.gene_num - 1)

    def random_select(self, with_id=False):
        if with_id:
            id = random.randint(0, len(self.gene) - 1)
            return self.gene[id], id
        else:
            return self.gene[random.randint(0, len(self.gene) - 1)]

    def tournament_selection(self, tournsize=2):
        """
        Current Issue:
        It may result in the loss of diversity
        """
        key = selTournamentFeature(list(enumerate(self.coef)), 1, tournsize=tournsize)
        return self.gene[key[0][0]]

    def best_gene(self):
        best_index = max(range(len(self.gene)), key=lambda x: np.abs(self.coef)[x])
        return self.gene[best_index]

    def worst_gene(self):
        worst_index = min(range(len(self.gene)), key=lambda x: np.abs(self.coef)[x])
        return self.gene[worst_index]

    def replace_worst_gene(self, gene):
        worst_index = min(range(len(self.gene)), key=lambda x: np.abs(self.coef)[x])
        self.gene[worst_index] = gene

    def softmax_selection(self, reverse=False, temperature=1 / 20):
        if temperature == 0:
            weight = self.coef
        else:
            if reverse:
                weight = softmax(-(1 / temperature) * self.coef)
            else:
                weight = softmax((1 / temperature) * self.coef)
        if weight.sum() == 0:
            # If there is an error in the feature importance vector
            index = np.random.choice(np.arange(len(self.gene)), 1)
        else:
            weight = weight / weight.sum()
            index = np.random.choice(np.arange(len(self.gene)), 1, p=weight)
        return self.gene[index[0]]

    def weighted_selection(self, reverse=False):
        weight = np.abs(self.coef)
        p = weight / np.sum(weight)
        if reverse:
            p[p == 0] = p[p > 0].min()
            p = 1 / p
            index = np.random.choice(np.arange(len(weight)), p=p / p.sum())
        else:
            index = np.random.choice(np.arange(len(weight)), p=p)
        return self.gene[index]

    def replace_weight_gene_inverse(self, gene):
        index = np.random.choice(np.arange(len(self.gene)), 1, p=softmax(-1 * np.abs(self.coef)))
        self.gene[index] = gene

    def deterministic_select(self):
        weight = np.abs(self.coef)
        return self.gene[np.argmax(-weight)]

    def __len__(self):
        return sum([len(g) for g in self.gene])

    def __str__(self):
        return str([str(g) for g in self.gene])

    def __repr__(self):
        return str([str(g) for g in self.gene])


def multiple_gene_evaluation(compiled_genes, x):
    result = []
    for gene in compiled_genes:
        result.append(gene(*x))
    return result


def multiple_gene_initialization(cls, generator, gene_num, **kwargs):
    return cls(generator, gene_num, **kwargs)


def multiple_gene_compile(expr: MultipleGeneGP, pset):
    """
    """
    gene_compiled = []
    if hasattr(expr, 'active_gene_num') and expr.active_gene_num > 0:
        # some genes may not be activated
        for gene in expr.gene[:expr.active_gene_num]:
            gene_compiled.append(compile(gene, pset))
    else:
        for gene in expr.gene:
            gene_compiled.append(compile(gene, pset))
    return gene_compiled


def pearson_check(x, y):
    if np.abs(pearsonr(x, y)[0]) < 0.95:
        return True
    else:
        return False


def cxOnePoint_multiple_gene_novelty(ind1: MultipleGeneGP, ind2: MultipleGeneGP, test_func):
    a = ind1.random_select()
    b = ind2.random_select()
    x, y = test_func(a), test_func(b)
    while True:
        cxOnePoint(a, b)
        z, w = test_func(a), test_func(b)
        if pearson_check(x, z) and pearson_check(y, z) and pearson_check(x, w) and pearson_check(y, w):
            break
    return ind1, ind2


def cxOnePoint_multiple_gene_diversity(ind1: MultipleGeneGP, ind2: MultipleGeneGP, visited_features: set):
    while True:
        a = ind1.random_select()
        b = ind2.random_select()
        cxOnePoint(a, b)
        if individual_to_tuple(a) in visited_features or individual_to_tuple(b) in visited_features:
            continue
    return ind1, ind2


def cxOnePoint_multiple_gene(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    if ind1.lgp_mode == True:
        gene, id = ind1.random_select(with_id=True)
        cxOnePoint(gene, ind2.gene[id])
    else:
        cxOnePoint(ind1.random_select(), ind2.random_select())
    return ind1, ind2


def cxOnePoint_multiple_gene_tournament(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    """
    A feature importance aware crossover operator based on the tournament selection
    """
    cxOnePoint(ind1.tournament_selection(), ind2.tournament_selection())
    return ind1, ind2


def cxOnePoint_multiple_gene_SC_Fixed(ind1: MultipleGeneGP, ind2: MultipleGeneGP, temperature=1 / 20):
    """
    self-competitive crossover operator
    Using the worst individual as the base individual and migrate a portion of useful materials from well-behaved one
    """
    ind1_copy = copy.deepcopy(ind1.softmax_selection(temperature=temperature))
    ind2_copy = copy.deepcopy(ind2.softmax_selection(temperature=temperature))
    cxOnePoint(ind1.softmax_selection(reverse=True, temperature=temperature), ind2_copy)
    cxOnePoint(ind2.softmax_selection(reverse=True, temperature=temperature), ind1_copy)
    return ind1, ind2


def selTournamentGenePool(coef_list, tournsize=7):
    individuals = [i for i in range(len(coef_list))]
    aspirants = selRandom(individuals, tournsize)
    return max(aspirants, key=lambda x: coef_list[x])


def selRouletteGenePool(coef_list):
    individuals = [i for i in range(len(coef_list))]
    s_inds = sorted(individuals, key=lambda x: coef_list[x], reverse=True)
    sum_fits = sum(coef_list[ind] for ind in individuals)
    u = random.random() * sum_fits
    sum_ = 0
    for ind in s_inds:
        sum_ += coef_list[ind]
        if sum_ >= u:
            return ind
    raise Exception('No solution found!')


def cxOnePoint_multiple_gene_pool(ind1: MultipleGeneGP, ind2: MultipleGeneGP,
                                  remove_zero_features=False):
    """
    Pay attention to remove redundant features and irrelevant features.
    """
    ind1_list, ind2_list = [], []
    ind1_set, ind2_set = set(), set()
    while len(ind1_list) < len(ind1.gene) or len(ind2_list) < len(ind2.gene):
        coef_list = np.array(list(ind1.coef) + list(ind2.coef))
        gene_list = ind1.gene + ind2.gene
        if remove_zero_features:
            # Remove completely irrelevant features
            gene_list = list(itertools.compress(ind1.gene + ind2.gene, coef_list > 0))
            coef_list = coef_list[coef_list > 0]
        a, b = selTournamentGenePool(coef_list), selTournamentGenePool(coef_list)
        if (a, b) in ind1_set:
            # Remove potential redundant features
            continue
        else:
            ind1_set.add((a, b))
        a, b = copy.deepcopy(gene_list[a]), copy.deepcopy(gene_list[b])
        cxOnePoint(a, b)
        if str(a) not in ind1_set and len(ind1_list) < len(ind1.gene):
            ind1_list.append(a)
            ind1_set.add(str(a))
        if str(b) not in ind2_set and len(ind2_list) < len(ind2.gene):
            ind2_list.append(b)
            ind2_set.add(str(b))
    ind1.gene = ind1_list
    ind2.gene = ind2_list
    return ind1, ind2


def cxOnePoint_multiple_gene_SC(ind1: MultipleGeneGP, ind2: MultipleGeneGP, temperature=1 / 20):
    """
    self-competitive crossover operator
    Using the worst individual as the base individual and migrate a portion of useful materials from well-behaved one
    """
    cxOnePoint(ind1.softmax_selection(reverse=True, temperature=temperature),
               copy.deepcopy(ind2.softmax_selection(temperature=temperature)))
    cxOnePoint(ind2.softmax_selection(reverse=True, temperature=temperature),
               copy.deepcopy(ind1.softmax_selection(temperature=temperature)))
    return ind1, ind2


def cxOnePoint_multiple_gene_same_index(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    # Crossover on the same location
    index = random.randint(0, ind1.gene_num - 1)
    cxOnePoint(ind1.gene[index], ind2.gene[index])
    return ind1, ind2


def cxOnePoint_multiple_gene_biased(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    cxOnePointLeafBiased(ind1.random_select(), ind2.random_select(), 0.1)
    return ind1, ind2


def cxOnePoint_multiple_gene_same_weight(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    # Only cross features with same weight index
    index = random.randint(0, ind1.gene_num - 1)
    cxOnePoint(ind1.gene[np.argsort(ind1.coef)[index]], ind2.gene[np.argsort(ind2.coef)[index]])
    return ind1, ind2


def cxOnePoint_multiple_gene_best(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    cxOnePoint(ind1.best_gene(), ind2.best_gene())
    return ind1, ind2


def cxOnePoint_multiple_gene_worst(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    cxOnePoint(ind1.worst_gene(), ind2.worst_gene())
    return ind1, ind2


# extract all useful features from individuals
def extract_features(ind: MultipleGeneGP, useful=True, threshold=None):
    results = []
    if threshold is not None:
        mean_coef = threshold * np.mean(ind.coef)
    else:
        mean_coef = np.mean(ind.coef)
    for i, x in zip(ind.gene, ind.coef):
        if useful and x >= mean_coef:
            results.append(i)
        if not useful and x < mean_coef:
            results.append(i)
    return results


def feature_to_tuple(x):
    return tuple(a.name for a in x)


def cxOnePoint_multiple_gene_threshold(ind1: MultipleGeneGP, ind2: MultipleGeneGP, mutation: Callable,
                                       cross_pb: float, threshold: float):
    # extract all useful features from individuals for crossover
    useful_features_a, useful_features_b = extract_features(ind1, threshold=threshold), \
                                           extract_features(ind2, threshold=threshold)
    # replace useless features with useful features
    all_features = set(feature_to_tuple(x) for x in extract_features(ind1, False, threshold) +
                       extract_features(ind2, False, threshold))

    # print('Count Useful Features', len(useful_features_a) + len(useful_features_b))
    # print('Count Useless Features', len(all_features))

    def replace_useless_features(ind):
        # generated_features = set(feature_to_tuple(x) for x in extract_features(ind))
        generated_features = set()
        # replace useless feature with useful features
        for i in range(len(ind.gene)):
            variation = True
            # check whether a feature is a useless feature
            while feature_to_tuple(ind.gene[i]) in all_features or variation:
                variation = False
                a, b = random.choice(useful_features_a), random.choice(useful_features_b)
                # if random.random() < (len(useful_features_a) + len(useful_features_b)) / all_features:
                if random.random() < cross_pb:
                    gene1, gene2 = cxOnePoint(copy.deepcopy(a), copy.deepcopy(b))
                else:
                    gene1, gene2 = mutation(copy.deepcopy(a))[0], mutation(copy.deepcopy(b))[0]
                # ensure the generated features are not redundant features
                l = list(filter(lambda x: (feature_to_tuple(x) not in all_features)
                                          and
                                          (feature_to_tuple(x) not in generated_features)
                                , [gene1, gene2]))
                if len(l) > 0:
                    ind.gene[i] = random.choice(l)
                    generated_features.add(feature_to_tuple(ind.gene[i]))

    replace_useless_features(ind1)
    replace_useless_features(ind2)
    return ind1, ind2


def cxOnePoint_multiple_gene_cross(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    gene1 = copy.deepcopy(ind1.best_gene())
    gene2 = copy.deepcopy(ind2.best_gene())
    gene1, gene2 = cxOnePoint(gene1, gene2)
    ind1.replace_worst_gene(gene1)
    ind2.replace_worst_gene(gene2)
    return ind1, ind2


def cxOnePoint_all_gene(ind1: MultipleGeneGP, ind2: MultipleGeneGP, permutation=False):
    if permutation:
        a_list = np.random.permutation(np.arange(0, len(ind1.gene)))
        b_list = np.random.permutation(np.arange(0, len(ind2.gene)))
    else:
        a_list = np.arange(0, len(ind1.gene))
        b_list = np.arange(0, len(ind2.gene))
    for a, b in zip(a_list, b_list):
        cxOnePoint(ind1.gene[a], ind2.gene[b])
    return ind1, ind2


def cxOnePoint_all_gene_with_importance_probability(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    # Each tree has a probability to be varied
    probability = (ind1.coef + ind2.coef) / 2
    a_list = np.arange(0, len(ind1.gene))
    b_list = np.arange(0, len(ind2.gene))
    for i, a, b in zip(range(len(a_list)), a_list, b_list):
        if random.random() < probability[i]:
            cxOnePoint(ind1.gene[a], ind2.gene[b])
    return ind1, ind2


def cxOnePoint_all_gene_with_probability(ind1: MultipleGeneGP, ind2: MultipleGeneGP, probability):
    # Each tree has a probability to be varied
    a_list = np.arange(0, len(ind1.gene))
    b_list = np.arange(0, len(ind2.gene))
    for a, b in zip(a_list, b_list):
        if random.random() < probability:
            cxOnePoint(ind1.gene[a], ind2.gene[b])
            # cxOnePointLeafBiased(ind1.gene[a], ind2.gene[b], 0.1)
    return ind1, ind2


def mutUniform_multiple_gene_with_probability(individual: MultipleGeneGP, expr, pset, probability):
    for g in individual.gene:
        if random.random() < probability:
            mutUniform(g, expr, pset)
    return individual,


# automated crossover operator?
# 1. feature importance

def mutUniform_multiple_gene(individual: MultipleGeneGP, expr, pset):
    if isinstance(pset, MultiplePrimitiveSet):
        gene, id = individual.random_select(with_id=True)
        mutUniform(gene, expr, pset.pset_list[id])
    else:
        mutUniform(individual.random_select(), expr, pset)
    return individual,


def mutShrink_multiple_gene(individual: MultipleGeneGP, expr, pset):
    if random.random() < 0.5:
        mutUniform(individual.random_select(), expr, pset)
    else:
        mutShrink(individual.random_select())
    return individual,


# transformer-based mutation
def mutUniform_multiple_gene_transformer(individual: MultipleGeneGP, expr, pset,
                                         condition_probability: Callable[[], float],
                                         transformer: TransformerTool):
    if random.random() > condition_probability():
        return mutUniform_multiple_gene(individual, expr, pset)
    else:
        ind = transformer.sample(1)
        individual.replace_worst_gene(PrimitiveTree(ind[0]))
        return individual,


def mutUniform_multiple_gene_worst(individual: MultipleGeneGP, expr, pset):
    mutUniform(individual.worst_gene(), expr, pset)
    return individual,


def mutUniform_multiple_gene_threshold(individual: MultipleGeneGP, expr, pset):
    s = sum([1 - c for c in individual.coef])
    for g, c in zip(individual.gene, individual.coef):
        if random.random() < (1 - c) / s:
            mutUniform(individual.worst_gene(), expr, pset)
    return individual,


def mutUniform_multiple_gene_with_prob(individual: MultipleGeneGP, expr, pset, terminal_probs, primitive_probs):
    root_individual = individual
    individual = individual.random_select()
    index = random.randrange(len(individual))
    slice_ = individual.searchSubtree(index)
    type_ = individual[index].ret
    individual[slice_] = expr(pset=pset, type_=type_, terminal_probs=terminal_probs, primitive_probs=primitive_probs)
    return root_individual,


def mutInsert_multiple_gene(individual: MultipleGeneGP, pset):
    mutInsert(individual.random_select(), pset)
    return individual,


def cxOnePoint_multiple_gene_weighted(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    # Potential issue: This operator may overly cross important features and ignore useless features
    cxOnePoint(ind1.weighted_selection(), ind2.weighted_selection())
    return ind1, ind2


def mutWeight_multiple_gene(individual: MultipleGeneGP, expr, pset, threshold_ratio=0.2):
    good_features, threshold = construct_feature_pools([individual], True, threshold_ratio=threshold_ratio)

    def replaces_features(ind: MultipleGeneGP):
        for i, c in enumerate(ind.coef):
            positive = False
            if (positive and c >= threshold) or (not positive and c < threshold):
                new_features = mutUniform(copy.deepcopy(random.choice(good_features)), expr, pset)
                ind.gene[i] = new_features[0]

    replaces_features(individual)
    return individual,


def construct_feature_pools(pop, positive, threshold_ratio=0.2,
                            good_features_threshold=None):
    # positive: get all important features
    # negative: get all unimportant features
    good_features = []
    good_features_str = set()
    if good_features_threshold == None:
        threshold = np.quantile([ind.coef for ind in pop], threshold_ratio)
    elif good_features_threshold == 'mean':
        threshold = np.mean([ind.coef for ind in pop])
    else:
        # threshold for good features
        threshold = np.quantile([ind.coef for ind in pop], good_features_threshold)

    def add_features(ind):
        for c, x in zip(ind.coef, ind.gene):
            if (positive and c >= threshold) or (not positive and c < threshold):
                if str(x) in good_features_str:
                    continue
                # assign a fitness value for each feature
                x.fitness = c
                good_features.append(x)
                good_features_str.add(str(x))

    for ind in pop:
        add_features(ind)

    # calculate the threshold for crossover
    threshold = np.quantile([ind.coef for ind in pop], threshold_ratio)
    return good_features, threshold


def feature_crossover_cross(ind1, ind2, threshold_ratio):
    pop = [ind1, ind2]
    good_features, threshold = construct_feature_pools(pop, True, threshold_ratio=threshold_ratio)
    new_pop = []
    for ind in pop:
        ind = cxOnePoint_multiple_gene_weight_plus(ind, good_features, threshold, False)
        new_pop.append(ind)
    return new_pop


def feature_crossover_cross_global(ind1, ind2, regressor):
    pop = [ind1, ind2]
    new_pop = []
    for ind in pop:
        ind = cxOnePoint_multiple_gene_weight_plus(ind, regressor.good_features, regressor.cx_threshold, False)
        new_pop.append(ind)
    return new_pop


def feature_mutation_global(individual: MultipleGeneGP, expr, pset, regressor):
    threshold = regressor.cx_threshold

    def replaces_features(ind: MultipleGeneGP):
        for i, c in enumerate(ind.coef):
            if c < threshold:
                new_features = mutUniform(copy.deepcopy(random.choice(regressor.good_features)), expr, pset)
                ind.gene[i] = new_features[0]

    replaces_features(individual)
    return individual,


def pool_based_mutation(individual: MultipleGeneGP, expr, pset, regressor, pearson_selection=False,
                        feature_evaluation=None, tournament_size=0):
    # construct features from the pool of good features
    # in addition to that, we force new features are not equivalent to old features and useless features
    def replaces_features(ind: MultipleGeneGP):
        for i, c in enumerate(ind.coef):
            if str(ind.gene[i]) in regressor.generated_features or c < regressor.cx_threshold:
                # new_features = copy.deepcopy(ind.gene[i])
                while True:
                    def feature_selection():
                        if tournament_size == 0:
                            return random.choice(regressor.good_features)
                        else:
                            return selTournament(regressor.good_features, tournament_size, 1)[0]

                    if random.random() < 0.2:
                        # mutation (in order to deal with the case of infinite loop)
                        new_features = mutUniform(copy.deepcopy(feature_selection()),
                                                  expr, pset)[0]
                    else:
                        # crossover
                        new_features = cxOnePoint(copy.deepcopy(feature_selection()),
                                                  copy.deepcopy(feature_selection()))
                        new_features = random.choice(new_features)
                    regressor.repetitive_feature_count[-1] += 1
                    if not str(new_features) in regressor.generated_features:
                        # Pre-selection by Pearson correlation to ensure the synthesized feature is useful
                        # However, such a process might be misleading
                        if pearson_selection:
                            func = compile(new_features, pset)
                            Yp = result_calculation([func], regressor.X[:20], False).flatten()
                            if np.abs(pearsonr(Yp, regressor.y[:20])[0]) <= 0.05:
                                # useless features
                                regressor.generated_features.add(str(new_features))
                            else:
                                break
                        elif feature_evaluation != None:
                            # using semantic diversity when generating new features
                            y = feature_evaluation(new_features)
                            if not y in regressor.generated_features:
                                break
                        else:
                            break
                ind.gene[i] = new_features
            # regressor.generated_features.add(str(new_features))

    replaces_features(individual)
    return individual,


def feature_crossover(ind1, ind2, positive, threshold_ratio):
    pop = [ind1, ind2]
    good_features, threshold = construct_feature_pools(pop, positive, threshold_ratio=threshold_ratio)
    new_pop = []
    for ind in pop:
        ind = cxOnePoint_multiple_gene_weight_plus(ind, good_features, threshold, positive)
        new_pop.append(ind)
    return new_pop


def cxOnePoint_multiple_gene_weight_plus(ind: MultipleGeneGP, good_features, threshold, positive):
    def replaces_features(ind: MultipleGeneGP):
        for i, c in enumerate(ind.coef):
            if (positive and c >= threshold) or (not positive and c < threshold):
                new_features = cxOnePoint(copy.deepcopy(random.choice(good_features)),
                                          copy.deepcopy(random.choice(good_features)))
                ind.gene[i] = random.choice(new_features)

    replaces_features(ind)
    return ind


def mutUniform_multiple_gene_weighted(individual: MultipleGeneGP, expr, pset):
    mutUniform(individual.weighted_selection(), expr, pset)
    return individual,


def cxOnePoint_multiple_gene_deterministic(ind1: MultipleGeneGP, ind2: MultipleGeneGP):
    cxOnePoint(ind1.deterministic_select(), ind2.deterministic_select())
    return ind1, ind2


def mutUniform_multiple_gene_deterministic(individual: MultipleGeneGP, expr, pset):
    mutUniform(individual.deterministic_select(), expr, pset)
    return individual,


def staticLimit(key, max_value, min_value):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            keep_inds = [copy.deepcopy(ind) for ind in args]
            new_inds = list(func(*args, **kwargs))
            for i, ind in enumerate(new_inds):
                if key(ind) > max_value or key(ind) < min_value:
                    new_inds[i] = random.choice(keep_inds)
            return new_inds

        return wrapper

    return decorator


def staticLimit_multiple_gene(key, max_value, min_value=0):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            keep_inds = [copy.deepcopy(ind.gene) for ind in args]
            new_inds = list(func(*args, **kwargs))
            for i, ind in enumerate(new_inds):
                ind: MultipleGeneGP
                for j, x in enumerate(ind.gene):
                    if callable(max_value):
                        height_limitation = max_value()
                    else:
                        height_limitation = max_value

                    if key(x) > height_limitation or key(x) < min_value:
                        # replace an unreasonable gene with a parent gene
                        # parent = keep_inds[i]
                        parent = random.choice(keep_inds)
                        gene = copy.deepcopy(parent[j])
                        ind.gene[j] = gene
                    assert key(ind.gene[j]) <= height_limitation
                    assert key(ind.gene[j]) >= min_value
            return new_inds

        return wrapper

    return decorator


def result_calculation(func, data, original_features):
    result = multiple_gene_evaluation(func, data.T)
    result = result_post_process(result, data, original_features)
    return result


def result_post_process(result, data, original_features):
    # some post process step
    result = quick_fill(result, data)
    if original_features:
        result = np.concatenate([np.array(result).T, data], axis=1)
    else:
        result = np.array(result).T
    return result


def quick_fill(result, data):
    for i in range(len(result)):
        yp = result[i]
        if not isinstance(yp, np.ndarray):
            yp = np.full(len(data), 0)
        elif yp.size == 1:
            yp = np.full(len(data), yp)
        result[i] = yp
    result = np.nan_to_num(result, posinf=0, neginf=0)
    return result


def genHalfAndHalf_with_prob(pset, min_, max_, terminal_probs, primitive_probs, type_=None):
    method = random.choice((genGrow_with_prob, genFull_with_prob))
    return method(pset, min_, max_, terminal_probs, primitive_probs, type_)


def genGrow_with_prob(pset, min_, max_, terminal_probs, primitive_probs, type_=None):
    def condition(height, depth):
        """Expression generation stops when the depth is equal to height
        or when it is randomly determined that a node should be a terminal.
        """
        return depth == height or \
               (depth >= min_ and random.random() < pset.terminalRatio)

    return generate_with_prob(pset, min_, max_, condition, terminal_probs, primitive_probs, type_)


def genFull_with_prob(pset, min_, max_, terminal_probs, primitive_probs, type_=None, sample_type=None):
    def condition(height, depth):
        """Expression generation stops when the depth is equal to height."""
        return depth == height

    return generate_with_prob(pset, min_, max_, condition, terminal_probs, primitive_probs, type_, sample_type)


def generate_with_prob(pset, min_, max_, condition, terminal_probs, primitive_probs, type_=None, sample_type=None):
    if type_ is None:
        type_ = pset.ret
    expr = []
    height = random.randint(min_, max_)
    stack = [(0, type_)]
    while len(stack) != 0:
        depth, type_ = stack.pop()
        if condition(height, depth):
            try:
                if sample_type == 'Dirichlet':
                    cat_prob = np.random.dirichlet(terminal_probs.flatten())
                    term = pset.terminals[type_][np.argmax(cat_prob)]
                else:
                    probability = terminal_probs.flatten()
                    probability = probability[:len(pset.terminals[type_])]
                    probability = probability / np.sum(probability)
                    term = np.random.choice(pset.terminals[type_], p=probability)
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The gp.generate function tried to add " \
                                 "a terminal of type '%s', but there is " \
                                 "none available." % (type_,)).with_traceback(traceback)
            if isclass(term):
                term = term()
            expr.append(term)
        else:
            try:
                prim = np.random.choice(pset.primitives[type_], p=primitive_probs.flatten())
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The gp.generate function tried to add " \
                                 "a primitive of type '%s', but there is " \
                                 "none available." % (type_,)).with_traceback(traceback)
            expr.append(prim)
            for arg in reversed(prim.args):
                stack.append((depth + 1, arg))
    return expr


class MultiplePrimitiveSet(PrimitiveSet):

    def __init__(self, name, arity, prefix="ARG"):
        super().__init__(name, arity, prefix)
        self.pset_list = []



import random

from sklearn.datasets import load_boston, load_diabetes
from torch.utils.tensorboard import SummaryWriter

import numpy as np
import torch
import math
from deap import gp, base
from scipy.stats import pearsonr
from sklearn.preprocessing import LabelEncoder, MinMaxScaler, StandardScaler
from torch import nn, Tensor
from torch.nn.utils.rnn import pad_sequence

from evolutionary_forest.component.evaluation import quick_evaluate
from evolutionary_forest.component.primitives import analytical_quotient
from evolutionary_forest.utils import reset_random
from experiment.experiment_dataset import get_dataset

device = 'cuda' if torch.cuda.is_available() else 'cpu'

reset_random(0)


class MutationTransformer(nn.Module):
    def __init__(
        self,
        embedding_class=16,
        embedding_size=64,
        dim_feedforward=64,
        num_layers=2,
        nhead=1,
        dropout=0.1,
        padding_idx=None):
        super().__init__()
        encoder_layer = nn.TransformerEncoderLayer(d_model=embedding_size, nhead=nhead, dropout=dropout,
                                                   dim_feedforward=dim_feedforward)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=num_layers)
        self.embedding_class = embedding_class
        self.embedding = nn.Embedding(embedding_class, embedding_size, padding_idx=padding_idx)

    def forward(self, batch_x, batch_context):
        torch_input_embedding = self.embedding(batch_x)
        torch_input_embedding = torch.cat([torch_input_embedding, batch_context], dim=1)
        torch_output = self.transformer_encoder.forward(torch.swapdims(torch_input_embedding, 0, 1))
        candidate_primitives = self.embedding(
            torch.IntTensor([i for i in range(self.embedding_class)]).to(device)
        )
        torch_output = torch_output[:number_of_primitives, :, :] @ candidate_primitives.T
        return torch.swapdims(torch_output, 0, 1)


"""
add(a,b)->multiply(a,b)
sin(a)->cos(a)
"""


def tuple_to_individual(pset, t):
    result = []
    for node in t:
        result.append(pset.mapping[node])
    tree = gp.PrimitiveTree(result)
    return tree


def model_training(full_dataset, X,
                   parameters=None):
    pset, primitive_encoder = get_primitive_set(X)
    writer = SummaryWriter('result')

    # Split training data
    train_size = int(0.8 * len(full_dataset))
    test_size = len(full_dataset) - train_size
    training_data, test_dataset = torch.utils.data.random_split(full_dataset, [train_size, test_size])
    # training_data = full_dataset[test_size:]
    # test_dataset = full_dataset[:test_size]
    torch_input, torch_context, torch_target = training_data_process(training_data, primitive_encoder)
    torch_input_test, torch_context_test, torch_target_test = training_data_process(test_dataset, primitive_encoder)
    number_of_primitives = len(primitive_encoder.classes_)
    model = MutationTransformer(embedding_class=number_of_primitives + 1,
                                padding_idx=number_of_primitives,
                                **parameters).to(device)
    loss_fn = nn.CrossEntropyLoss()
    for iter in range(20000):
        # X is a torch Variable
        permutation = torch.randperm(torch_input.size()[0]).to(device)
        batch_size = 128

        # Training in a mini-batch mode
        for i in range(0, torch_input.size()[0], batch_size):
            indices = permutation[i:i + batch_size]
            batch_x, batch_y = torch_input[indices], torch_target[indices]
            optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
            optimizer.zero_grad()
            torch_output = model.forward(batch_x, torch_context[indices])
            loss = loss_fn(torch.swapdims(torch_output, 1, 2), batch_y)
            loss.backward()
            optimizer.step()
        # Evaluate on the training set
        model.eval()
        torch_output_train = model.forward(torch_input, torch_context)
        training_accuracy_nn, training_accuracy_random = calculate_accuracy(number_of_primitives,
                                                                            pset,
                                                                            training_data,
                                                                            torch_input,
                                                                            torch_output_train,
                                                                            torch_context,
                                                                            torch_target,
                                                                            primitive_encoder,
                                                                            testing_flag=False)
        # Evaluate on the test set
        torch_output_test = model.forward(torch_input_test, torch_context_test)
        testing_accuracy_nn, testing_accuracy_random = calculate_accuracy(number_of_primitives,
                                                                          pset,
                                                                          test_dataset,
                                                                          torch_input_test,
                                                                          torch_output_test,
                                                                          torch_context_test,
                                                                          torch_target_test,
                                                                          primitive_encoder)
        model.train()

        writer.add_scalar('Accuracy/Train-Transformer', training_accuracy_nn, iter)
        writer.add_scalar('Accuracy/Train-Random', training_accuracy_random, iter)
        writer.add_scalar('Accuracy/Val-Transformer', testing_accuracy_nn, iter)
        writer.add_scalar('Accuracy/Val-Random', testing_accuracy_random, iter)

    writer.close()
    return training_accuracy, testing_accuracy


def calculate_accuracy(number_of_primitives,
                       pset,
                       all_test_data,
                       torch_input_test,
                       torch_output_test,
                       torch_context_test,
                       torch_target_test,
                       primitive_encoder,
                       testing_flag=True):
    all_inds = []
    for t in all_test_data:
        ind = tuple_to_individual(pset, t[0])
        all_inds.append(ind)

    context_data = torch_context_test.detach().numpy()
    context_y = context_data[:, 0]
    context_x = context_data[:, 1:]

    # Perform mutation
    valid_mutation = 0
    valid_random_mutation = 0
    all_mutation = 0
    label = np.argmax(torch_output_test.detach().numpy(), axis=2)
    # label = torch_target_test.detach().numpy()
    for id, item in enumerate(label):
        ind = all_inds[id]
        # Make the new individual feasible
        current_terminal_id = 0
        for node_id, a in enumerate(ind):
            if isinstance(a, gp.Terminal):
                # do replacement
                ind[node_id] = pset.mapping[f'ARG{current_terminal_id}']
                current_terminal_id += 1
        old_pearson = pearsonr(quick_evaluate(ind, pset, context_x[id].T), context_y[id])[0]

        # Try to perform Transformer-based Mutation
        current_pos = 0
        new_nodes = []
        for node_id, a in enumerate(ind):
            if isinstance(a, gp.Primitive):
                # do replacement
                if item[current_pos] < len(primitive_encoder.classes_):
                    # ensure it is a valid replacement
                    new_node = pset.mapping[primitive_encoder.classes_[item[current_pos]]]
                    new_nodes.append(new_node)
                current_pos += 1
            else:
                new_nodes.append(a)
        ind = gp.PrimitiveTree(new_nodes)

        # Evaluate the performance of new features
        new_pearson = pearsonr(quick_evaluate(ind, pset, context_x[id].T), context_y[id])[0]
        if np.abs(new_pearson) > 10 * np.abs(old_pearson):
            valid_mutation += 1

        # Try to perform Random Mutation
        new_nodes = []
        for node_id, a in enumerate(ind):
            if isinstance(a, gp.Primitive):
                # ensure it is a valid replacement
                all_primitives = list(filter(lambda x: x.arity == a.arity, pset.primitives[object]))
                new_node = random.choice(all_primitives)
                # print('Old vs New', ind[node_id].name, new_node.name)
                new_nodes.append(new_node)
            else:
                new_nodes.append(a)
        ind = gp.PrimitiveTree(new_nodes)

        # Evaluate the performance of new features
        new_pearson = pearsonr(quick_evaluate(ind, pset, context_x[id].T), context_y[id])[0]
        if np.abs(new_pearson) > 10 * np.abs(old_pearson):
            valid_random_mutation += 1
        all_mutation += 1
    print('Accuracy of Mutation', valid_mutation / all_mutation)
    print('Accuracy of Random Mutation', valid_random_mutation / all_mutation)

    for c in range(torch_output_test.shape[1]):
        # Number of different predictions
        testing = 'Testing' if testing_flag else 'Training'
        print(
            'Position', c,
            '%s Accuracy' % testing,
            torch.sum((torch.argmax(torch_output_test[:, c], dim=1) == torch_target_test[:, c])
                      [torch_target_test[:, c] != number_of_primitives]) /
            torch.sum(torch_target_test[:, c] != number_of_primitives)
        )
    total_accuracy = torch.sum((torch.argmax(torch_output_test, dim=2) == torch_target_test)
                               [torch_target_test != number_of_primitives]) / \
                     torch.sum(torch_target_test != number_of_primitives)

    return valid_mutation / all_mutation, valid_random_mutation / all_mutation


def training_data_process(training_data, primitive_encoder):
    # Preparing training data
    torch_input = []
    torch_target = []
    torch_context = []
    for t in training_data:
        t_a = list(filter(lambda x: 'ARG' not in x, t[0]))
        t_b = list(filter(lambda x: 'ARG' not in x, t[1]))
        s_a = primitive_encoder.transform(t_a)
        s_b = primitive_encoder.transform(t_b)
        torch_input.append(torch.IntTensor(s_a))
        context = torch.FloatTensor(t[2])
        if number_of_terminals + 1 - context.shape[1] > 0:
            context = torch.cat([context,
                                 torch.zeros((context.shape[0], number_of_terminals + 1 - context.shape[1]))],
                                dim=1)
        torch_context.append(context.T)
        torch_target.append(torch.LongTensor(s_b))
    torch_input = pad_sequence(torch_input, padding_value=len(primitive_encoder.classes_)).T
    torch_target = pad_sequence(torch_target, padding_value=len(primitive_encoder.classes_)).T
    torch_context = torch.stack(torch_context)

    # input = torch_context.numpy()
    # np.concatenate([torch_input.numpy(), torch_target.numpy()], axis=1)
    # lgbm_score = cross_validate(LGBMClassifier(), input.reshape(input.shape[0], -1)[:, :80],
    #                             torch_target.numpy()[:, 0],
    #                             return_train_score=True)
    # print(lgbm_score['train_score'])
    # print(lgbm_score['test_score'])

    return torch_input.to(device), torch_context.to(device), torch_target.to(device)


def get_primitive_set(X):
    pset = gp.PrimitiveSet("MAIN", X.shape[1])
    pset.addPrimitive(np.add, 2)
    pset.addPrimitive(np.subtract, 2)
    pset.addPrimitive(np.multiply, 2)
    pset.addPrimitive(analytical_quotient, 2)
    pset.addPrimitive(np.maximum, 2)
    pset.addPrimitive(np.minimum, 2)
    # pset.addPrimitive(np.sin, 1)
    # pset.addPrimitive(np.cos, 1)
    # pset.addPrimitive(_protected_sqrt, 1)
    primitive_encoder = LabelEncoder()
    primitive_encoder.fit([p.name for p in filter(lambda x: isinstance(x, gp.Primitive), pset.primitives[object])])
    return pset, primitive_encoder


def training_data_generation(X, y):
    # Generating training data
    pset, _ = get_primitive_set(X)
    toolbox = base.Toolbox()
    toolbox.register("expr", gp.genHalfAndHalf, pset=pset, min_=1, max_=height)
    training_data = []
    hash_list = set()
    for _ in range(50000):
        id = np.random.randint(0, len(X), embedding_size)
        y_sample = y[id]
        tree = toolbox.expr()

        result_a = quick_evaluate(tree, pset, X[id])
        tree_a = tuple(a.name for a in tree)

        # ensure the training data is useful
        while True:
            for node_id, node in enumerate(tree):
                if isinstance(node, gp.Primitive):
                    all_primitives = list(filter(lambda x: x.arity == node.arity, pset.primitives[object]))
                    new_node = random.choice(all_primitives)
                    tree[node_id] = new_node
            tree_b = tuple(a.name for a in tree)
            if tree_b != tree_a:
                break

        used_variable = []
        for node in tree:
            if isinstance(node, gp.Terminal):
                used_variable.append(int(node.name.replace('ARG', '')))
        assert len(used_variable) <= number_of_terminals
        result_b = quick_evaluate(tree, pset, X[id])
        used_data = X[id][:, list(used_variable)]
        assert used_data.shape[1] <= number_of_terminals
        used_data = np.concatenate((np.reshape(y[id], (-1, 1)), used_data), axis=1)
        if np.abs(pearsonr(result_b, y_sample)[0]) > 10 * np.abs(pearsonr(result_a, y_sample)[0]):
            # cross_validate(LinearRegression(), np.reshape(result_b, (-1, 1)), y_sample, return_train_score=True)
            # Avoid ambiguous label
            # hash_result = hash(tree_a)
            # if hash_result in hash_list:
            #     continue
            # else:
            #     hash_list.add(hash_result)
            # pearsonr(np.subtract(used_data[:,1],used_data[:,2]),used_data[:,0])[0]
            print(np.abs(pearsonr(result_a, y_sample)[0]), np.abs(pearsonr(result_b, y_sample)[0]))
            training_data.append((tree_a, tree_b, used_data))
        elif np.abs(pearsonr(result_a, y_sample)[0]) > 10 * np.abs(pearsonr(result_b, y_sample)[0]):
            print(np.abs(pearsonr(result_a, y_sample)[0]), np.abs(pearsonr(result_b, y_sample)[0]))
            training_data.append((tree_b, tree_a, used_data))
        else:
            pass
    print('Useful Training Data', len(training_data))
    return training_data


def get_dataset_simple():
    X, y = load_boston(return_X_y=True)
    # X, y, _ = get_dataset({
    #     'dataset': '604_fri_c4_500_10'
    # })
    # X, y = load_diabetes(return_X_y=True)
    X, y = np.array(X), np.array(y)
    X = StandardScaler().fit_transform(X)
    y = StandardScaler().fit_transform(y.reshape(-1, 1)).flatten()
    return X, y


if __name__ == '__main__':
    embedding_size = 64
    dim_feedforward = 32
    num_layers = 2
    nhead = 4
    dropout = 0.5
    height = 2
    number_of_primitives = 2 ** height - 1
    number_of_terminals = 2 ** height

    X, y = get_dataset_simple()
    training_data = training_data_generation(X, y)
    training_accuracy, testing_accuracy = model_training(training_data, X,
                                                         {
                                                             'embedding_size': embedding_size,
                                                             'dim_feedforward': dim_feedforward,
                                                             'num_layers': num_layers,
                                                             'nhead': nhead,
                                                             'dropout': dropout,
                                                         })



import math
import random
from functools import partial
from typing import Union, Callable

import numpy as np
from deap import base
from deap.tools import cxUniform, mutUniformInt
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering, DBSCAN
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, StandardScaler
from sklearn.tree import DecisionTreeRegressor

from evolutionary_forest.multigene_gp import result_calculation


class SpacePartition():
    def __init__(self):
        self.X: np.ndarray = None
        self.y: np.ndarray = None
        self.base_learner: str = None
        self.dynamic_partition: str = None
        self.current_gen: int = None
        self.n_gen: int = None
        self.ps_tree_ratio: float = None
        self.interleaving_period: int = None
        self.partition_model: BaseEstimator = None

    def partition_number_calculation(self):
        """
        Dynamically calculate the partition number based on training data
        :return:
        """
        if isinstance(self.partition_number, str) and 'Sample' in self.partition_number:
            min_samples_leaf = int(self.partition_number.split('-')[1])
            partition_number = int(self.partition_number.split('-')[2])
            partition_number = int(np.clip(math.floor(self.X.shape[0] / min_samples_leaf), 1, partition_number))
            self.partition_number = partition_number
            return partition_number
        else:
            assert isinstance(self.partition_number, int)
            return self.partition_number

    def partition_scheme_initialization(self) -> Union[np.ndarray, Callable, None]:
        """
        Initialize a space partition scheme
        """
        if self.base_learner == 'Soft-PLTree' or 'Fast-Soft-PLTree' in self.base_learner:
            if self.dynamic_partition == 'Self-Adaptive' or self.dynamic_partition is None:
                # Initialization with decision tree
                if isinstance(self.partition_number, str) and 'Sample' in self.partition_number:
                    min_samples_leaf = int(self.partition_number.split('-')[1])
                    partition_number = int(self.partition_number.split('-')[2])
                    dt = DecisionTreeRegressor(min_samples_leaf=min_samples_leaf,
                                               max_leaf_nodes=partition_number)
                    dt.fit(self.X, self.y)
                    self.partition_number = dt.tree_.n_leaves
                else:
                    dt = DecisionTreeRegressor(max_leaf_nodes=self.partition_number)
                    dt.fit(self.X, self.y)
                self.partition_model = dt
                partition_scheme = LabelEncoder().fit_transform(dt.apply(self.X))
                assert np.all(partition_scheme < self.partition_number)
            elif self.dynamic_partition == 'GA':
                self.partition_number = self.partition_number_calculation()

                def partition_scheme():
                    return np.random.randint(0, self.partition_number, len(self.X))
            else:
                raise Exception
        else:
            partition_scheme = None
        return partition_scheme

    def partition_scheme_updating(self):
        # Full evaluation after some iterations
        if (self.interleaving_period is None or self.interleaving_period == 0) \
            and self.current_gen > self.n_gen * (1 - self.ps_tree_ratio):
            if self.base_learner.startswith('Fast-'):
                self.base_learner = self.base_learner.replace('Fast-', '')

        if self.interleaving_period > 0:
            # Interleaving changing
            if self.current_gen % self.interleaving_period == 0:
                if not self.base_learner.startswith('Fast'):
                    self.base_learner = 'Fast-' + self.base_learner
            else:
                self.base_learner = self.base_learner.replace('Fast-', '')
            if self.verbose:
                print(self.current_gen, self.base_learner, self.interleaving_period)


def partition_scheme_varAnd(population, toolbox, cxpb, mutpb):
    offspring = [toolbox.clone(ind) for ind in population]
    # Apply crossover and mutation on the offspring
    for i in range(1, len(offspring), 2):
        if random.random() < cxpb:
            offspring[i - 1], offspring[i] = toolbox.mate(offspring[i - 1], offspring[i])
    for i in range(len(offspring)):
        if random.random() < mutpb:
            offspring[i], = toolbox.mutate(offspring[i])
    return offspring


def partition_scheme_toolbox(partition_num, cxpb=0.1, indpb=0.1):
    toolbox = base.Toolbox()
    # toolbox.register("mate", cxTwoPoint)
    toolbox.register("mate", cxUniform, indpb=cxpb)
    toolbox.register("mutate", partial(mutUniformInt, low=0, up=partition_num - 1, indpb=indpb))
    return toolbox



#!/usr/bin/env python

"""Tests for `evolutionary_forest` package."""

from numpy.testing import assert_almost_equal
from sklearn.datasets import make_regression
from sklearn.metrics import mean_squared_error

from evolutionary_forest.forest import EvolutionaryForestRegressor


def test_simple_data():
    X, y = make_regression(n_samples=100, n_features=5, n_informative=5)
    gp = EvolutionaryForestRegressor(max_height=8, normalize=True, select='AutomaticLexicase', boost_size=10, n_gen=2,
                                     gene_num=5, base_learner='Random-DT')
    gp.fit(X, y)
    assert_almost_equal(mean_squared_error(y, gp.predict(X)), 0)
    assert len(gp.hof) == 10
    assert len(gp.hof[0].gene) == 5



# this module is used for supporting feature importance based genetic programming
import random
import sys
from inspect import isclass

import numpy as np


def genGrow(pset, min_, max_, type_=None, terminal_prob=None):
    """Generate an expression where each leaf might have a different depth
    between *min* and *max*.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param type_: The type that should return the tree when called, when
                  :obj:`None` (default) the type of :pset: (pset.ret)
                  is assumed.
    :returns: A grown tree with leaves at possibly different depths.
    """

    def condition(height, depth):
        """Expression generation stops when the depth is equal to height
        or when it is randomly determined that a node should be a terminal.
        """
        return depth == height or \
               (depth >= min_ and random.random() < pset.terminalRatio)

    return generate(pset, min_, max_, condition, type_, terminal_prob=terminal_prob)


def genHalfAndHalf(pset, min_, max_, type_=None, terminal_prob=None):
    """Generate an expression with a PrimitiveSet *pset*.
    Half the time, the expression is generated with :func:`~deap.gp.genGrow`,
    the other half, the expression is generated with :func:`~deap.gp.genFull`.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param type_: The type that should return the tree when called, when
                  :obj:`None` (default) the type of :pset: (pset.ret)
                  is assumed.
    :returns: Either, a full or a grown tree.
    """
    method = random.choice((genGrow, genFull))
    return method(pset, min_, max_, type_, terminal_prob=terminal_prob)


def genFull(pset, min_, max_, type_=None, terminal_prob=None, primitive_prob=None):
    """Generate an expression where each leaf has the same depth
    between *min* and *max*.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param type_: The type that should return the tree when called, when
                  :obj:`None` (default) the type of :pset: (pset.ret)
                  is assumed.
    :returns: A full tree with all leaves at the same depth.
    """

    def condition(height, depth):
        """Expression generation stops when the depth is equal to height."""
        return depth == height

    return generate(pset, min_, max_, condition, type_,
                    terminal_prob=terminal_prob, primitive_prob=primitive_prob)


def generate(pset, min_, max_, condition, type_=None, terminal_prob=None, primitive_prob=None):
    """Generate a Tree as a list of list. The tree is build
    from the root to the leaves, and it stop growing when the
    condition is fulfilled.

    :param pset: Primitive set from which primitives are selected.
    :param min_: Minimum height of the produced trees.
    :param max_: Maximum Height of the produced trees.
    :param condition: The condition is a function that takes two arguments,
                      the height of the tree to build and the current
                      depth in the tree.
    :param type_: The type that should return the tree when called, when
                  :obj:`None` (default) the type of :pset: (pset.ret)
                  is assumed.
    :returns: A grown tree with leaves at possibly different depths
              depending on the condition function.
    """
    if type_ is None:
        type_ = pset.ret
    expr = []
    height = random.randint(min_, max_)
    stack = [(0, type_)]
    while len(stack) != 0:
        depth, type_ = stack.pop()
        if condition(height, depth):
            try:
                if terminal_prob is None:
                    term = random.choice(pset.terminals[type_])
                else:
                    term = np.random.choice(pset.terminals[type_], p=terminal_prob)
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The gp.generate function tried to add " \
                                 "a terminal of type '%s', but there is " \
                                 "none available." % (type_,)).with_traceback(traceback)
            except ValueError:
                raise
            if isclass(term):
                term = term()
            expr.append(term)
        else:
            try:
                if primitive_prob is None:
                    prim = random.choice(pset.primitives[type_])
                else:
                    prim = np.random.choice(pset.primitives[type_], p=primitive_prob)
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The gp.generate function tried to add " \
                                 "a primitive of type '%s', but there is " \
                                 "none available." % (type_,)).with_traceback(traceback)
            expr.append(prim)
            for arg in reversed(prim.args):
                stack.append((depth + 1, arg))
    return expr



import random

import numpy
import numpy as np
from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap.tools import cxTwoPoint
from sklearn.metrics import r2_score


def oob_pruning(oob, y_train):
    ensemble_size = len(oob)

    def oob_error(x):
        x = np.array(x).astype(bool)
        if (np.sum((oob[x] != 0), axis=0) == 0).any():
            return -10 * np.sum(np.sum((oob[x] != 0), axis=0) == 0),
        prediction = np.sum(oob[x], axis=0) / np.sum((oob[x] != 0), axis=0)
        if np.isnan(prediction).any():
            return -10 * np.sum(np.isnan(prediction)),
        oob_score = r2_score(y_train, prediction)
        return oob_score,

    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMax)

    toolbox = base.Toolbox()

    toolbox.register("attr_bool", random.randint, 0, 1)
    toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, n=ensemble_size)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)

    toolbox.register("evaluate", oob_error)
    toolbox.register("mate", cxTwoPoint)
    toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
    toolbox.register("select", tools.selTournament, tournsize=10)

    pop = toolbox.population(n=100)

    # Numpy equality function (operators.eq) between two arrays returns the
    # equality element wise, which raises an exception in the if similar()
    # check of the hall of fame. Using a different equality function like
    # numpy.array_equal or numpy.allclose solve this issue.
    hof = tools.HallOfFame(1, similar=numpy.array_equal)

    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", numpy.mean)
    stats.register("std", numpy.std)
    stats.register("min", numpy.min)
    stats.register("max", numpy.max)

    algorithms.eaSimple(pop, toolbox, cxpb=0.5, mutpb=0.2, ngen=50, stats=stats,
                        halloffame=hof, verbose=False)
    weight = np.array((hof[0]))
    return weight



import math
import operator
from collections import defaultdict
from functools import partial
from itertools import chain, compress
from operator import eq

import numpy as np
from deap.tools import HallOfFame
from sklearn.linear_model import RidgeCV, LinearRegression
from sklearn.metrics import r2_score
from sklearn.preprocessing import OneHotEncoder
from sklearn.tree import DecisionTreeRegressor

from evolutionary_forest.component.evaluation import quick_result_calculation
from evolutionary_forest.component.primitives import individual_to_tuple
from evolutionary_forest.component.subset_selection import EnsembleSelectionADE


class StrictlyImprovementHOF(HallOfFame):

    def update(self, population):
        population = list(filter(lambda x: (not hasattr(x, 'parent_fitness')) or
                                           np.all(x.fitness.wvalues[0] > np.array(x.parent_fitness)),
                                 population))
        super().update(population)


class GeneralizationHOF(HallOfFame):
    def __init__(self, X, y, pset, maxsize=1, verbose=False):
        super().__init__(maxsize)
        self.X = X
        self.y = y
        self.pset = pset
        self.verbose = verbose

    def update(self, population):
        best_ind = sorted(population, key=lambda x: x.fitness.wvalues)[-1]
        X_hat = quick_result_calculation(best_ind.gene, self.pset, self.X)
        lr_r2_scores = []
        for _ in range(5):
            id = np.random.permutation(np.arange(0, len(self.X)))
            y = self.y[id]
            lr = LinearRegression()
            lr_r2_score = r2_score(y, lr.fit(X_hat, y).predict(X_hat))
            lr_r2_scores.append(lr_r2_score)
        best_ind.complexity_score = np.mean(lr_r2_scores)
        if len(self) == 0:
            super().update(population)
        else:
            if best_ind.fitness.wvalues[0] - best_ind.complexity_score > \
                self[0].fitness.wvalues[0] - self[0].complexity_score:
                super().update(population)
            else:
                if self.verbose:
                    print('Complexity Score', best_ind.complexity_score, self[0].complexity_score)


class LexicaseHOF(HallOfFame):
    """
    Select all individuals with one case speciality
    """

    def __init__(self):
        HallOfFame.__init__(self, None)

    def update(self, population):
        for p in population:
            self.insert(p)

        hofer_arr: np.ndarray = None
        for i, hofer in enumerate(self):
            if hofer_arr is None:
                hofer_arr = np.array(-1 * hofer.case_values).reshape(1, -1)
            else:
                hofer_arr = np.concatenate([hofer_arr, np.array(-1 * hofer.case_values).reshape(1, -1)])
        max_hof = np.max(hofer_arr, axis=0)

        del_ind = []
        max_value = np.full_like(max_hof, -np.inf)
        for index, x in enumerate(self):
            fitness_wvalues = np.array(-1 * x.case_values)
            if np.any(fitness_wvalues >= max_hof) \
                and np.any(fitness_wvalues > max_value):
                loc = np.where(fitness_wvalues > max_value)
                max_value[loc] = fitness_wvalues[loc]
                continue
            # determine the deletion individual list
            del_ind.append(index)

        for i in reversed(del_ind):
            self.remove(i)


class RandomObjectiveHOF(HallOfFame):
    """
    An archive which updating based on random projected fitness values
    """

    def update(self, population):
        for ind in population:
            if len(self) == 0 and self.maxsize != 0:
                # Working on an empty hall of fame is problematic for the
                # "for else"
                self.insert(population[0])
                continue
            if np.any(ind.case_values < self[-1].case_values) or len(self) < self.maxsize:
                for hofer in self:
                    # Loop through the hall of fame to check for any
                    # similar individual
                    if self.similar(ind, hofer):
                        break
                else:
                    # The individual is unique and strictly better than
                    # the worst
                    if len(self) >= self.maxsize:
                        self.remove(-1)
                    self.insert(ind)


"""
There are two places need to consider data imbalance
"""


class EnsembleSelectionHallOfFame(HallOfFame):
    def __init__(self, maxsize, label, similar=eq, multitask=False, diversity_ratio=0,
                 unique=True, bagging_iteration=1, loss_function=None, inner_sampling=0,
                 outer_sampling=0.5, limited_sample=False, initial_size=5):
        """
        Three strategies:
        1. Selection with Replacement
        2. Sorted Ensemble Initialization
        3. Bagged Ensemble Selection

        unique: Selection without bootstrapping, and thus avoiding weighted selection
        """
        super().__init__(maxsize, similar)
        self.label = label
        self.categories = None
        self.novelty_weight = None
        self.unique = unique
        self.ensemble_weight = defaultdict(int)
        self.initial_size = initial_size
        self.bagging_iteration = bagging_iteration
        self.instances_num = 0
        self.verbose = False
        self.class_weight = None
        self.task_type = 'Regression'
        self.loss_function = loss_function
        self.multitask = multitask
        self.diversity_ratio = diversity_ratio
        self.inner_sampling = inner_sampling
        self.outer_sampling = outer_sampling
        self.limited_sample = limited_sample

    def mse(self, prediction):
        return (self.label - prediction) ** 2

    def cross_entropy(self, prediction):
        eps = 1e-15
        return -1 * self.label * np.log(np.clip(prediction, eps, 1 - eps))

    def zero_one(self, prediction):
        return (np.argmax(self.label, axis=1) != np.argmax(prediction, axis=1))

    def hinge_loss(self, prediction):
        return np.mean(np.clip(self.label - prediction, 0, None))

    def loss(self, prediction):
        # Both classification and regression use MSE as the ensemble selection metric
        # The reason is due to an additive model need to use MSE as the loss function
        if self.task_type == 'Regression':
            return self.mse(prediction)
        else:
            if self.loss_function == 'Hinge':
                return self.hinge_loss(prediction)
            elif self.loss_function == 'CrossEntropy':
                return self.cross_entropy(prediction)
            elif self.loss_function == 'ZeroOne':
                return self.zero_one(prediction)
            elif self.loss_function == 'MSE':
                return self.mse(prediction)
            else:
                raise Exception

    def ensemble_selection(self, population):
        new_inds = []
        new_ind_tuples = defaultdict(int)
        instances_num = self.instances_num
        assert instances_num > 0
        if self.categories == None:
            # regression
            sum_prediction = np.zeros(instances_num)
        else:
            # classification
            sum_prediction = np.zeros((instances_num, self.categories))

        all_inds = population
        population_size = len(all_inds)
        current_error = np.inf
        # we need to define a variable to record
        selection_count = 0

        if self.loss_function == 'ZeroOne':
            # ZeroOne loss is beneficial for decision tree
            for ind in population:
                # ind.predicted_values = (ind.predicted_values == np.max(ind.predicted_values, axis=0)).astype(int)
                argmax = np.argmax(ind.predicted_values, axis=1)
                ind.predicted_values[:] = 0
                ind.predicted_values[np.arange(0, len(ind.predicted_values)), argmax] = 1

        def error_calculation(x, diversity=False):
            prediction = ((x.predicted_values + sum_prediction) / (selection_count + 1))
            if self.class_weight is not None:
                loss = self.loss(prediction)
                if len(loss.shape) == 1:
                    error = np.sum(self.class_weight.flatten() * loss)
                elif len(loss.shape) == 2:
                    error = np.sum(self.class_weight * loss)
                else:
                    raise Exception
            else:
                error = np.sum(self.loss(prediction))
            if diversity:
                error -= self.diversity_ratio * \
                         np.sum((x.predicted_values - sum_prediction / selection_count) ** 2)
            return error

        """
        目前依然需要确定多样性丧失的问题根源
        """
        while selection_count < self.maxsize and len(all_inds) > 0:
            if selection_count == 0:
                if isinstance(self, GreedySelectionHallOfFame):
                    # calculate loss summation
                    if self.loss_function in ['ZeroOne', 'CrossEntropy', 'MSE']:
                        errors = [error_calculation(x) for x in all_inds]
                    else:
                        errors = [np.sum(x.case_values[:instances_num]) for x in all_inds]
                    # For multitask optimization, we may need to consider them respectively
                    if self.multitask and all_inds[0].base_model != None:
                        initial_individuals = self.initial_size * 2
                        args = np.argsort(errors)
                        all_base_models = {ind.base_model for ind in all_inds}
                        inds = []
                        for m in all_base_models:
                            inds.extend(list(filter(lambda id: all_inds[id].base_model == m, args))
                                        [:self.initial_size])
                    else:
                        initial_individuals = self.initial_size
                    inds = np.argsort(errors)[:initial_individuals]
                    for ind in inds:
                        ind = all_inds[ind]
                        sum_prediction += ind.predicted_values
                        selection_count += 1
                        # using tuple to determine which individuals are selected
                        ind_tuple = individual_to_tuple(ind)
                        new_ind_tuples[ind_tuple] += 1
                        if self.unique:
                            # single instance selection
                            new_inds.append(ind)
                            all_inds.remove(ind)
                            self.ensemble_weight[ind_tuple] += 1
                        else:
                            # allow sample with replacement
                            if ind_tuple not in self.ensemble_weight:
                                new_inds.append(ind)
                                self.ensemble_weight[ind_tuple] = 1
                            else:
                                self.ensemble_weight[ind_tuple] += 1
                    if self.unique:
                        assert len(all_inds) < population_size, f'{len(all_inds)} {population_size}'
                    assert selection_count == initial_individuals, selection_count
                    # update current error (first iteration)
                    prediction = (sum_prediction / initial_individuals)
                    if self.class_weight is not None:
                        current_error = np.sum(self.class_weight * self.loss(prediction))
                    else:
                        current_error = np.sum(self.loss(prediction))
                    continue
                else:
                    ind = min(all_inds, key=lambda x: np.sum(x.case_values[:instances_num]))
            else:
                if isinstance(self, DREPHallOfFame):
                    if self.class_weight is not None or self.task_type == 'Classification':
                        raise Exception
                    # First select individuals according to diversity
                    elitist = sorted(all_inds,
                                     key=lambda x: -1 * np.sum((x.predicted_values -
                                                                sum_prediction / selection_count) ** 2)) \
                        [:math.ceil(self.r * len(all_inds))]
                    # Then select individuals according to accuracy
                    ind = min(elitist, key=lambda x: np.sum(x.case_values[:instances_num]))
                elif isinstance(self, GreedySelectionHallOfFame):
                    # Select based on loss reduction
                    # list(map(lambda x: (x.base_model, error_calculation(x)), all_inds))
                    if self.inner_sampling > 0:
                        selected_index = np.random.random(len(all_inds)) < self.inner_sampling
                        if np.sum(selected_index) == 0:
                            # if all zero, then random set one position to be true
                            selected_index[np.random.randint(len(all_inds))] = True
                        ind_pool = list(compress(all_inds, selected_index))
                    else:
                        ind_pool = all_inds
                    if self.diversity_ratio > 0:
                        ind = min(ind_pool, key=partial(error_calculation, diversity=True))
                    else:
                        ind = min(ind_pool, key=error_calculation)
                else:
                    raise Exception

            # common early stop strategy
            early_stop = True
            skip_selection = True
            if error_calculation(ind) < current_error:
                current_error = error_calculation(ind)
            else:
                # warning: different order will not produce different results
                all_inds.remove(ind)
                if isinstance(self, GreedySelectionHallOfFame):
                    # If the selection strategy is Greedy Selection, then we don't need to check further more
                    break
                if skip_selection:
                    continue
                if early_stop:
                    break

            sum_prediction += ind.predicted_values
            selection_count += 1
            # using tuple to determine which individuals are selected
            ind_tuple = individual_to_tuple(ind)
            new_ind_tuples[ind_tuple] += 1
            if self.unique:
                # single instance selection
                new_inds.append(ind)
                all_inds.remove(ind)
                self.ensemble_weight[ind_tuple] += 1
            else:
                if ind_tuple not in self.ensemble_weight:
                    new_inds.append(ind)
                    self.ensemble_weight[ind_tuple] = 1
                else:
                    self.ensemble_weight[ind_tuple] += 1
                    if self.limited_sample and new_ind_tuples[ind_tuple] >= 5:
                        # limit the number of same model each round
                        all_inds.remove(ind)
        # assert_almost_equal(np.sum(sum_prediction / selection_count, axis=1), 1)
        if self.class_weight is not None:
            loss = np.mean(self.class_weight * self.loss(sum_prediction / selection_count))
        else:
            loss = np.mean(self.loss(sum_prediction / selection_count))
        return new_inds, loss

    def update(self, population):
        self.ensemble_weight.clear()
        self.instances_num = len(population[0].predicted_values)
        if self.bagging_iteration > 1:
            new_inds = []
            new_inds_set = set()
            for _ in range(self.bagging_iteration):
                probability = self.outer_sampling
                all_inds = list(chain(population, self))
                selected_index = np.random.random(len(all_inds)) < probability
                temp_inds, _ = self.ensemble_selection(list(compress(all_inds, selected_index)))
                for ind in temp_inds:
                    ind_tuple = individual_to_tuple(ind)
                    if ind_tuple not in new_inds_set:
                        new_inds.append(ind)
                        new_inds_set.add(ind_tuple)
            # [(self.ensemble_weight.get(individual_to_tuple(a), 0), a) for a in all_inds]
            # print('All Ensemble Weight', self.ensemble_weight.values())
            assert len(self.ensemble_weight.keys()) == len(new_inds_set)
        else:
            new_inds, _ = self.ensemble_selection(population)
        if self.verbose:
            print('Ensemble Size', len(new_inds))
        self.clear()
        for x in new_inds:
            self.insert(x)


class DREPHallOfFame(EnsembleSelectionHallOfFame):
    # A hall of fame based on DREP algorithm
    def update(self, population):
        paradigm = []
        for r in np.arange(0.2, 0.5 + 0.01, 0.1):
            self.r = r
            paradigm.append(self.ensemble_selection(population))
        # print([x[1] for x in paradigm])
        new_inds, _ = min(paradigm, key=lambda x: x[1])
        super(DREPHallOfFame, self).update(new_inds)


class NoveltyHallOfFame(EnsembleSelectionHallOfFame):
    pass


class GreedySelectionHallOfFame(EnsembleSelectionHallOfFame):
    pass


class HybridHallOfFame(HallOfFame):
    """
    A hall of fame tailored for hybrid base models,
    i.e., retain best models for each class of base models
    """

    def __init__(self, maxsize, similar=operator.eq):
        super().__init__(maxsize, similar=similar)
        hof_num = 2
        self.hofs = [HallOfFame(maxsize // hof_num) for i in range(hof_num)]

    def update(self, population):
        for cls, hof in zip([RidgeCV, DecisionTreeRegressor], self.hofs):
            hof.update(list(filter(lambda x: isinstance(x.pipe['Ridge'], cls), population)))

    def __len__(self):
        return len(list(chain.from_iterable(self.hofs)))

    def __getitem__(self, i):
        return list(chain.from_iterable(self.hofs))[i]

    def __iter__(self):
        return iter(chain.from_iterable(self.hofs))

    def __reversed__(self):
        return reversed(list(chain.from_iterable(self.hofs)))


class OOBHallOfFame(HallOfFame):
    """
    A hall of fame based on OOB error
    """

    def __init__(self, X, y, toolbox, maxsize, similar=operator.eq):
        super().__init__(maxsize, similar)
        self.X = X
        self.y = y
        self.toolbox = toolbox

    # @timeit
    def update(self, population):
        if len(self.items) < self.maxsize:
            super().update(population)
        else:
            classes = population[0].pipe['Ridge'].classes_
            n_classes = population[0].pipe['Ridge'].n_classes_

            # generate one-hot label
            encoded_label = OneHotEncoder(sparse=False).fit_transform(self.y.reshape(-1, 1))
            assert np.all(classes[np.argmax(encoded_label, axis=1)] == self.y)
            assert encoded_label.shape[1] == n_classes

            current_pop = list(self.items) + population
            prediction = np.full((len(current_pop), len(self.y), n_classes), np.nan)
            for i, x in enumerate(current_pop):
                index = x.out_of_bag
                prediction[i][index] = x.oob_prediction

            def evaluation_func(x):
                # calculate the OOB error for a given selection scheme
                x_index = np.argsort(x)[:self.maxsize]
                return np.sum((np.nanmean(prediction[x_index], axis=0) - encoded_label) ** 2)

            # generate an initial vector based on fitness values
            initial_vector = np.zeros(len(current_pop))
            initial_vector[np.argsort(-1 * np.array([p.fitness.wvalues[0] for p in current_pop]))[:self.maxsize]] = -1
            e = EnsembleSelectionADE(len(current_pop), 20, 100, evaluation_function=evaluation_func,
                                     verbose=False, de_algorithm='JSO')
            result = e.run(initial_vector=initial_vector)
            selected_index = np.argsort(result)[:self.maxsize]

            self.clear()
            for s in selected_index:
                self.insert(current_pop[s])
            assert len(self.items) == self.maxsize, len(self.items)


class BootstrapHallOfFame(HallOfFame):
    """
    A hall of fame based on different optimal targets
    """

    def __init__(self, x_train, maxsize, similar=operator.eq):
        super().__init__(maxsize, similar)
        self.x_train = x_train
        sample_indices = np.arange(self.x_train.shape[0])
        bootstrap_indices = np.random.choice(sample_indices,
                                             size=(maxsize, sample_indices.shape[0]),
                                             replace=True)
        self.bootstrap_indices = bootstrap_indices
        self.items = [None for _ in range(maxsize)]
        self.items_score = [np.inf for _ in range(maxsize)]

    def update(self, population):
        for p in population:
            for i, s in enumerate(self.bootstrap_indices):
                score = np.sum(p.case_values[s])
                if score < self.items_score[i]:
                    self.items[i] = p
                    self.items_score[i] = score



import random
import sys
from inspect import isclass

import numpy as np
from deap.tools import HallOfFame


def data_to_tensor(data: np.ndarray, sample: tuple):
    indices = np.random.randint(0, len(data), sample)
    return indices, data[indices]


def hof_to_text(hof: HallOfFame, indices):
    case_values = np.array([x.case_values for x in hof])
    text = []
    for ix in indices:
        # only select top-1 feature
        ind = hof[np.argmin(np.sum(case_values[:, ix], axis=1))]
        index = np.argmax(ind.coef)
        x = ind.gene[index]
        s = ' '.join(tuple(a.name for a in x))
        text.append(s)
    return text


def expr_generate(feature: list, pset, type_=None):
    feature = list(filter(lambda x: x not in ['<unk>', '<pad>', '<bos>', '<eos>'], feature))
    expr = []
    stack = [(0, type_)]
    while len(stack) != 0:
        depth, type_ = stack.pop()
        if len(feature) > 0:
            x = feature.pop(0)
            if x in pset.mapping:
                term = pset.mapping[x]
                # constant terminals
                if isclass(term):
                    term = term()
                # primitives
                if hasattr(term, 'args'):
                    for arg in reversed(term.args):
                        stack.append((depth + 1, arg))
                expr.append(term)
            else:
                # constant terminals
                term = pset.terminals[object][-1]()
                term.value = int(x)
                term.name = str(term.value)
        else:
            # add some end terminals
            try:
                term = random.choice(pset.terminals[type_])
            except IndexError:
                _, _, traceback = sys.exc_info()
                raise IndexError("The gp.generate function tried to add " \
                                 "a terminal of type '%s', but there is " \
                                 "none available." % (type_,)).with_traceback(traceback)
            if isclass(term):
                term = term()
            expr.append(term)
    return expr


class TransformerTool():
    def __init__(self, X, y, hof, pset):
        self.transformer = None
        self.pset = pset
        self.X = np.concatenate([X, y.reshape(-1, 1)], axis=1)
        self.hof = hof
        self.embedding_size = 32

    def train(self):
        from .translation_transformer import TranslationTransformer
        from torchtext.data.datasets_utils import _RawTextIterableDataset
        X, hof = self.X, self.hof
        indices, X = data_to_tensor(X, sample=(len(hof), self.embedding_size))
        X = np.swapaxes(X, 1, 2).astype(np.float32)
        text = hof_to_text(hof, indices)

        def data_iterator(split='train'):
            iterator = _RawTextIterableDataset('GP', len(hof), zip(list(X).__iter__(), text.__iter__()))
            return iterator

        epochs = 500
        if self.transformer == None:
            t = TranslationTransformer(data_iterator, EMB_SIZE=self.embedding_size)
            t.train(epochs)
            self.transformer = t
        else:
            t = self.transformer
            t.get_data_iterator = data_iterator
            t.train(epochs)

    def sample(self, num):
        X = self.X
        _, X_test = data_to_tensor(X, sample=(num, self.embedding_size))
        X_test = np.swapaxes(X_test, 1, 2).astype(np.float32)
        samples = [expr_generate(self.transformer.generate(np.expand_dims(x, axis=1)), self.pset, object) for x in
                   X_test]
        return samples



import numpy as np
from deap.gp import Primitive, Terminal
from deap.tools import HallOfFame
from mlxtend.evaluate import feature_importance_permutation

from evolutionary_forest.component.evaluation import quick_result_calculation, get_cv_splitter


class EstimationOfDistribution():
    def __init__(self):
        self.pset = None
        self.mutation_scheme: str = None
        self.hof: HallOfFame = None
        self.primitive_prob: np.ndarray = None
        self.terminal_prob: np.ndarray = None

    def permutation_importance_calculation(self, X, Y, individual):
        # The correct way is to calculate the terminal importance based on the test data
        estimators = individual.estimators
        kcv = get_cv_splitter(estimators[0]['Ridge'], self.cv)
        all_importance_values = []
        for id, index in enumerate(kcv.split(X, Y)):
            def prediction_function(X):
                # quickly calculate permutation importance
                Yp = quick_result_calculation(individual.gene, self.pset, X, self.original_features)
                return estimators[id].predict(Yp)

            train_index, test_index = index
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = Y[train_index], Y[test_index]
            # get permutation importance
            importance_value = feature_importance_permutation(X_test, y_test, prediction_function, 'r2')[0]
            all_importance_values.append(importance_value)
        all_importance_values = np.mean(all_importance_values, axis=0)
        all_importance_values = np.clip(all_importance_values, 0, None)
        individual.terminal_importance_values = all_importance_values

    def update_frequency_count(self, best_pop, factor, importance_weight):
        # count the frequency of primitives and terminals
        primitive_dict = {}
        for k in self.pset.primitives.keys():
            primitive_dict.update({v.name: k for k, v in enumerate(self.pset.primitives[k])})
        terminal_dict = {}
        for k in self.pset.terminals.keys():
            terminal_dict.update({v.name: k for k, v in enumerate(self.pset.terminals[k])})
        for h in best_pop:
            for importance, g in zip(h.coef, h.gene):
                for x in g:
                    if importance_weight:
                        weight = importance
                    else:
                        weight = 1
                    weight *= factor
                    if isinstance(x, Primitive):
                        self.primitive_prob_count[primitive_dict[x.name]] += weight
                    elif isinstance(x, Terminal):
                        if x.name not in terminal_dict:
                            # constant
                            self.terminal_prob_count[-1] += weight
                        else:
                            self.terminal_prob_count[terminal_dict[x.name]] += weight
                    else:
                        raise Exception

    def frequency_counting(self, importance_weight=True):
        if self.mutation_scheme == 'EDA-Terminal-PMI':
            # Permutation importance based probability estimation (Slow!)
            for h in self.hof:
                if not hasattr(h, 'terminal_importance_values'):
                    self.permutation_importance_calculation(self.X, self.y, h)
            self.primitive_prob_count = np.ones(self.pset.prims_count)
            self.terminal_prob_count = np.mean([x.terminal_importance_values for x in self.hof], axis=0)
        else:
            ### Feature Importance based on frequency
            self.primitive_prob_count = np.ones(self.pset.prims_count)
            self.terminal_prob_count = np.ones(self.pset.terms_count)
            if self.hof == None:
                return
            if self.mutation_scheme == 'EDA-Terminal-Balanced':
                # Separate individuals to two groups, good ones and bad ones
                positive_sample = list(sorted(self.hof, key=lambda x: x.fitness.wvalues[0]))[len(self.hof) // 2:]
                negative_sample = list(sorted(self.hof, key=lambda x: x.fitness.wvalues[0]))[:len(self.hof) // 2]
                self.update_frequency_count(positive_sample, 1, importance_weight)
                self.update_frequency_count(negative_sample, -1, importance_weight)
                min_max_scaling = lambda v: (v - v.min())
                self.primitive_prob_count = min_max_scaling(self.primitive_prob_count) + 1
                self.terminal_prob_count = min_max_scaling(self.terminal_prob_count) + 1
            else:
                if 'Population' in self.mutation_scheme:
                    self.update_frequency_count(self.pop, 1, importance_weight)
                else:
                    self.update_frequency_count(self.hof, 1, importance_weight)

    def probability_sampling(self):
        # Sampling primitive and terminal nodes based on the probability matrix
        if self.mutation_scheme == 'probability-TS':
            self.primitive_prob[:] = np.random.dirichlet(self.primitive_prob_count)
            self.terminal_prob[:] = np.random.dirichlet(self.terminal_prob_count)
        elif self.mutation_scheme == 'EDA-Primitive':
            self.primitive_prob[:] = np.random.dirichlet(self.primitive_prob_count)
            self.terminal_prob /= np.sum(self.terminal_prob)
        elif self.mutation_scheme in ['EDA-Terminal']:
            self.primitive_prob /= np.sum(self.primitive_prob)
            self.terminal_prob += self.terminal_prob_count / np.sum(self.terminal_prob_count)
        elif self.mutation_scheme in ['EDA-Terminal-Balanced', 'EDA-Terminal-PMI']:
            self.primitive_prob[:] = self.primitive_prob_count / np.sum(self.primitive_prob_count)
            if self.mutation_scheme == 'EDA-Terminal-PMI':
                # add the probability of constant nodes
                terminal_prob = np.append(self.terminal_prob_count, np.mean(self.terminal_prob_count))
            else:
                terminal_prob = self.terminal_prob_count
            assert np.all(terminal_prob >= 0)
            if np.sum(terminal_prob) == 0:
                # When there are no important features
                self.terminal_prob[:] = np.full_like(terminal_prob, 1 / len(terminal_prob))
            else:
                self.terminal_prob[:] = terminal_prob / np.sum(terminal_prob)
        elif self.mutation_scheme.startswith('EDA-PM'):
            self.primitive_prob[:] = self.primitive_prob_count / np.sum(self.primitive_prob_count)
            self.terminal_prob[:] = self.terminal_prob_count / np.sum(self.terminal_prob_count)
            if self.verbose:
                print('primitive_prob', self.primitive_prob)
                print('terminal_prob', self.terminal_prob)
        elif self.mutation_scheme.startswith('EDA-Terminal-PM') or self.mutation_scheme in \
            ['weight-plus-threshold-EDA', 'EDA-Terminal-SameWeight', 'EDA-Terminal-PM-Tournament']:
            self.primitive_prob /= np.sum(self.primitive_prob)
            self.terminal_prob[:] = self.terminal_prob_count / np.sum(self.terminal_prob_count)
            ## Eliminate trivial terminals
            # self.terminal_prob[self.terminal_prob < np.mean(self.terminal_prob) / 10] = 0
            # self.terminal_prob[self.terminal_prob > 0] = 1
            # self.terminal_prob[:] = self.terminal_prob / np.sum(self.terminal_prob)
        else:
            raise Exception
        if self.verbose:
            print('terminal_prob', self.terminal_prob)



import copy
import os
import random
import time
from collections import defaultdict
from itertools import chain

import numpy as np
import pandas as pd
import seaborn as sns
import torch
from deap.gp import Primitive
from matplotlib import pyplot as plt
from numba import njit
from sympy import latex, parse_expr

from evolutionary_forest.component.primitives import individual_to_tuple


def get_important_features(parents):
    # Get important features
    genes = list(chain.from_iterable([x.gene for x in parents]))
    coefficients = list(chain.from_iterable([x.coef for x in parents]))
    hash_codes = list(chain.from_iterable([x.hash_result for x in parents]))
    results = dict()
    for g, c, h in zip(genes, coefficients, hash_codes):
        if h not in results:
            results[h] = [float(c), g]
        else:
            results[h][0] += float(c)
    features = list(sorted(results.items(), key=lambda item: -1 * item[1][0]))
    coefs = list(map(lambda x: x[1][0], features))
    features = list(map(lambda x: x[1][1], features))
    return features, coefs


def efficient_deepcopy(self, memo=None):
    if memo is None:
        memo = {}
    cls = self.__class__  # Extract the class of the object
    result = cls.__new__(cls)  # Create a new instance of the object based on extracted class
    memo[id(self)] = result
    for k, v in self.__dict__.items():
        if k in ('predicted_values', 'case_values', 'pipe', 'estimators'):
            continue
        # Copy over attributes by copying directly or in case of complex objects like lists
        # for example calling the `__deepcopy()__` method defined by them.
        # Thus recursively copying the whole tree of objects.
        setattr(result, k, copy.deepcopy(v, memo))
    return result


def get_feature_importance(regr, latex_version=True, fitness_weighted=False, mean_fitness=False,
                           ensemble_weighted=True):
    """
    :param regr: evolutionary forest
    :param latex_version: return simplified symbol, which is used for printing
    :param fitness_weighted: assign different weights to features based on fitness values
    :param mean_fitness: return mean feature importance instead of summative feature importance
    :return:
    """
    if mean_fitness:
        all_genes_map = defaultdict(list)
    else:
        all_genes_map = defaultdict(int)
    hash_dict = {}

    # Processing function
    if latex_version:
        latex_string = lambda g: latex(parse_expr(gene_to_string(g).replace("ARG", "X").replace("_", "-")),
                                       mul_symbol='dot')
        processing_code = lambda g: f'${latex_string(g)}$'
    else:
        # Generating a Python code fragment
        def code_generation(gene):
            code = str(gene)
            args = ",".join(arg for arg in regr.pset.arguments)
            code = "lambda {args}: {code}".format(args=args, code=code)
            return code

        processing_code = lambda g: f'{code_generation(g)}'

    for x in regr.hof:
        for o_g, h, c in zip(x.gene, x.hash_result, np.abs(x.coef)):
            # Taking the fitness of each model into consideration
            importance_value = c
            if fitness_weighted:
                importance_value = importance_value * x.fitness.wvalues[0]
            if ensemble_weighted and hasattr(regr.hof, 'ensemble_weight'):
                importance_value = importance_value * regr.hof.ensemble_weight[individual_to_tuple(x)]
            # Merge features with equivalent hash values
            if h not in hash_dict or len(o_g) < len(hash_dict[h]):
                hash_dict[h] = o_g
            # Deciding using summation or mean technique to calculate the importance value
            if mean_fitness:
                all_genes_map[h].append(importance_value)
            else:
                all_genes_map[h] += importance_value
    # Forming a dict
    all_genes_map = {
        processing_code(hash_dict[k]): all_genes_map[k] for k in all_genes_map.keys()
    }
    if mean_fitness:
        for k, v in all_genes_map.items():
            all_genes_map[k] = np.mean(v)
    feature_importance_dict = {k: v for k, v in sorted(all_genes_map.items(), key=lambda item: -item[1])}
    sum_value = np.sum([v for k, v in feature_importance_dict.items()])
    feature_importance_dict = {k: v / sum_value for k, v in feature_importance_dict.items()}
    return feature_importance_dict


def select_top_features(code_importance_dict, ratio=None):
    """
    :param code_importance_dict: feature importance dict
    :param ratio: only select top features
    :return:
    """
    if ratio == None:
        mean_importance = np.mean(list(code_importance_dict.values()))
    else:
        mean_importance = np.quantile(list(code_importance_dict.values()), ratio)
    features = list(map(lambda kv: kv[0], filter(lambda kv: kv[1] >= mean_importance, code_importance_dict.items())))
    return features


def feature_append(regr, X, features, only_new_features=False):
    if only_new_features:
        data = []
    else:
        data = [X]
    for f in features:
        func = eval(f, regr.pset.context)
        features = func(*X.T)
        if isinstance(features, np.ndarray) and len(features) == len(X):
            data.append(features.reshape(-1, 1))
    transformed_features = np.concatenate(data, axis=1)
    # Fix outliers (In extreme cases, some functions will produce unexpected results)
    transformed_features = np.nan_to_num(transformed_features, posinf=0, neginf=0)
    return transformed_features


def plot_feature_importance(feature_importance_dict, save_fig=False):
    names, importance = list(feature_importance_dict.keys()), list(feature_importance_dict.values())
    # Create arrays from feature importance and feature names
    feature_importance = np.array(importance)
    feature_names = names

    # Create a DataFrame using a Dictionary
    data = {'feature_names': feature_names, 'feature_importance': feature_importance}
    fi_df = pd.DataFrame(data)

    # Sort the DataFrame in order decreasing feature importance
    fi_df.sort_values(by=['feature_importance'], ascending=False, inplace=True)

    # Define size of bar plot
    plt.figure(figsize=(12, 8))
    sns.set(style="white", font_scale=1.5)
    # Plot Seaborn bar chart
    sns.barplot(x=fi_df['feature_importance'][:15], y=fi_df['feature_names'][:15], palette='bone')

    # Add chart labels
    plt.xlabel('Feature Importance')
    plt.ylabel('Feature Name')
    plt.tight_layout()
    if save_fig:
        plt.savefig('result/feature_importance.png', format='png')
        plt.savefig('result/feature_importance.eps', format='eps')
    plt.show()


infix_map = {
    'Add': '+',
    'Sub': '-',
    'Mul': '*',
    'Div': '/',
}


def gene_to_string(gene):
    string = ""
    stack = []
    for node in gene:
        stack.append((node, []))
        while len(stack[-1][1]) == stack[-1][0].arity:
            prim, args = stack.pop()
            # string = prim.format(*args)
            if type(prim) is Primitive:
                string = '('
                if prim.name == 'AQ':
                    string += f'{args[0]}/sqrt(1+{args[1]}*{args[1]})'
                elif prim.name == 'Log':
                    string += f'log(1+Abs({args[0]}))'
                elif prim.name == 'Sqrt':
                    string += f'sqrt(Abs({args[0]}))'
                elif prim.name == 'Max':
                    string += f'Max({args[0]}, {args[1]})'
                elif prim.name == 'Min':
                    string += f'Min({args[0]}, {args[1]})'
                elif prim.name == 'Neg':
                    string += f'-{args[0]}'
                elif prim.name not in infix_map:
                    string += prim.format(*args)
                else:
                    string += args[0]
                    for a in args[1:]:
                        string += f'{infix_map[prim.name]}{a}'
                string += ')'
            else:
                string = prim.name
            if len(stack) == 0:
                break  # If stack is empty, all nodes should have been seen
            stack[-1][1].append(string)
    return string


@njit(cache=True)
def pairwise_data(x, y):
    data = []
    label = []
    for a, va in zip(x, y):
        for b, vb in zip(x, y):
            data.append(a - b)
            # data.append(np.concatenate((a, b)))
            # label.append(va > vb)
            label.append(va > vb)
    return data, label


def pairwise_data_np(x, y):
    data, label = pairwise_data(x, y)
    return np.array(data), np.array(label)


@njit(cache=True)
def calculate_value(x):
    data = []
    count = np.round(np.sqrt(len(x)))
    for i in range(0, len(x), count):
        s = 0
        for j in range(i, i + count):
            s += x[j]
        s = s / count
        data.append(s)
    return data


def get_activations(clf, X):
    # get hidden layer outputs
    hidden_layer_sizes = clf.hidden_layer_sizes
    if not hasattr(hidden_layer_sizes, "__iter__"):
        hidden_layer_sizes = [hidden_layer_sizes]
    hidden_layer_sizes = list(hidden_layer_sizes)
    layer_units = [X.shape[1]] + hidden_layer_sizes + \
                  [clf.n_outputs_]
    activations = [X]
    for i in range(clf.n_layers_ - 1):
        activations.append(np.empty((X.shape[0],
                                     layer_units[i + 1])))
    clf._forward_pass(activations)
    return activations


if __name__ == '__main__':
    print(parse_expr('log(1)'))
    print(latex(parse_expr('X1*X2'), mul_symbol='dot'))


@njit(cache=True)
def numba_seed(a):
    random.seed(a)
    np.random.seed(a)


def reset_random(s):
    random.seed(s)
    np.random.seed(s)
    numba_seed(s)
    torch.manual_seed(s)
    os.environ['PYTHONHASHSEED'] = str(s)


def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    values, weights -- Numpy ndarrays with the same shape.
    """
    average = np.average(values, weights=weights, axis=0)
    # Fast and numerically precise:
    variance = np.average((values - average) ** 2, weights=weights, axis=0)
    return (average, np.sqrt(variance))


def save_array(array):
    save_object(array, f'np_{time.time()}.pkl')


def save_object(obj, path):
    import pickle
    with open(path, 'wb') as file:
        pickle.dump(obj, file)


def load_object(path):
    import pickle
    with open(path, 'rb') as file:
        obj = pickle.load(file)
        return obj
